"""
FIDO2 1.x API implementation for ProtonVPN session handling.

Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=no-name-in-module
from fido2.hid import CtapHidDevice
from fido2.client import (
    Fido2Client, PublicKeyCredentialRequestOptions as Options,
    UserInteraction as Fido2UserInteraction,
    AssertionSelection
)

from proton.session.api import Fido2AssertionParameters, Fido2Assertion


def create_client(device: CtapHidDevice,
                  origin: str,
                  user_interaction: Fido2UserInteraction) -> Fido2Client:
    """Create a FIDO2 client for the given device."""

    return Fido2Client(  # pylint: disable=unexpected-keyword-arg
        device,
        origin,
        user_interaction=user_interaction
    )


def create_options(assertion_parameters: Fido2AssertionParameters) -> Options:
    """Create a FIDO2 options for the given assertion parameters."""

    return Options(
        challenge=assertion_parameters.challenge,
        rp_id=assertion_parameters.rp_id,
        allow_credentials=assertion_parameters.allow_credentials,
        user_verification=assertion_parameters.user_verification
    )


def create_from_client_assertion(assertion: AssertionSelection) -> Fido2Assertion:
    """Create a FIDO2 assertion from the given client assertion."""

    response = assertion.get_response(0)
    return Fido2Assertion(
        client_data=bytes(response.client_data),
        authenticator_data=bytes(response.authenticator_data),
        signature=bytes(response.signature),
        credential_id=bytes(response.credential_id)
    )



"""
Wrapper over the NetworkManager client.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from concurrent.futures import Future
from threading import Thread, Lock
from typing import Callable, Optional

import gi
gi.require_version("NM", "1.0")  # noqa: required before importing NM module
# pylint: disable=wrong-import-position
from gi.repository import NM, GLib

from proton.vpn.connection.exceptions import VPNConnectionError

logger = logging.getLogger(__name__)


class NMClient:
    """
    Wrapper over the NetworkManager client.
    It also starts the GLib main loop used by the NetworkManager client.
    """
    _lock = Lock()
    _main_context = None
    _nm_client = None

    @classmethod
    def initialize_nm_client_singleton(cls):
        """
        Initializes the NetworkManager client singleton.

        If the singleton was initialized, this method will do nothing. However,
        if the singleton wasn't initialized it will initialize it, starting
        a new GLib MainLoop.

        A double-checked lock is used to avoid the possibility of multiple
        threads concurrently creating multiple instances of the NM client
        (with their own main loops).
        """
        if cls._nm_client:
            return

        with cls._lock:
            if not cls._nm_client:
                cls._initialize_nm_client_singleton()

    @classmethod
    def _initialize_nm_client_singleton(cls):
        cls._main_context = GLib.MainContext()
        cls._nm_client = NM.Client()
        # Setting daemon=True when creating the thread makes that this thread
        # exits abruptly when the python process exits. It would be better to
        # exit the thread running the main loop calling self._main_loop.quit().
        Thread(target=cls._run_main_loop, daemon=True).start()

        callback, future = cls.create_nmcli_callback(
            finish_method_name="new_finish"
        )

        def new_async():
            cls._assert_running_on_main_loop_thread()
            cls._nm_client.new_async(cancellable=None, callback=callback, user_data=None)

        cls._run_on_main_loop_thread(new_async)
        cls._nm_client = future.result()

    @classmethod
    def _run_main_loop(cls):
        main_loop = GLib.MainLoop(cls._main_context)
        cls._main_context.push_thread_default()
        main_loop.run()

    @classmethod
    def _assert_running_on_main_loop_thread(cls):
        """
        This method asserts that the thread running it is the one iterating
        GLib's main loop.

        It's useful to call this method at the beginning of any code block
        that's supposed to run in GLib's main loop, to avoid hard-to-debug
        issues.

        For more info:
        https://developer.gnome.org/documentation/tutorials/main-contexts.html#checking-threading
        """
        assert cls._main_context.is_owner()  # nosec B311, B101 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B101

    @classmethod
    def _run_on_main_loop_thread(cls, function):
        cls._main_context.invoke_full(priority=GLib.PRIORITY_DEFAULT, function=function)

    @classmethod
    def create_nmcli_callback(cls, finish_method_name: str) -> (Callable, Future):
        """Creates a callback for the NM client finish method and a Future that will
        resolve once the callback is called."""
        future = Future()
        future.set_running_or_notify_cancel()

        def callback(source_object, res, userdata):  # pylint: disable=unused-argument
            cls._assert_running_on_main_loop_thread()
            try:
                # On errors, according to the docs, the callback can be called
                # with source_object/res set to None.
                # https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/Client.html#NM.Client.new_async
                if not source_object or not res:

                    raise VPNConnectionError(
                        f"An unexpected error occurred initializing NMClient: "
                        f"source_object = {source_object}, res = {res}."
                    )

                result = getattr(source_object, finish_method_name)(res)

                # According to the docs, None is returned on errors
                # https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/Client.html#NM.Client.new_finish
                if not result:
                    raise VPNConnectionError(
                        "An unexpected error occurred initializing NMCLient"
                    )

                future.set_result(result)
            except BaseException as exc:  # pylint: disable=broad-except
                future.set_exception(exc)

        return callback, future

    def __init__(self):
        self.initialize_nm_client_singleton()

    def commit_changes_async(
            self, new_connection: NM.RemoteConnection
    ) -> Future:
        """
        Commits changes asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/RemoteConnection.html#NM.RemoteConnection.commit_changes_async
        :return: a Future to keep track of completion.
        """
        callback, future = self.create_nmcli_callback(
            finish_method_name="commit_changes_finish"
        )

        def commit_changes_async():
            self._assert_running_on_main_loop_thread()
            new_connection.commit_changes_async(
                True,
                None,
                callback,
                None
            )

        self._run_on_main_loop_thread(commit_changes_async)
        return future

    def add_connection_async(self, connection: NM.Connection) -> Future:
        """
        Adds a new connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/Client.html#NM.Client.add_connection_async
        :param connection: connection to be added.
        :return: a Future to keep track of completion.
        """
        callback, future = self.create_nmcli_callback(
            finish_method_name="add_connection_finish"
        )

        def add_connection_async():
            self._assert_running_on_main_loop_thread()
            self._nm_client.add_connection_async(
                connection=connection,
                save_to_disk=False,
                cancellable=None,
                callback=callback,
                user_data=None
            )

        self._run_on_main_loop_thread(add_connection_async)
        return future

    def start_connection_async(self, connection: NM.Connection) -> Future:
        """Starts a VPN connection asynchronously.
        :param connection: connection to be started.
        :return: Future to know when the connection has been started. Note that
        is just after the connection has started but before it is established.
        """
        callback, future = self.create_nmcli_callback(
            finish_method_name="activate_connection_finish"
        )

        def activate_connection_async():
            self._assert_running_on_main_loop_thread()
            self._nm_client.activate_connection_async(
                connection,
                None,
                None,
                None,
                callback,
                None
            )

        self._run_on_main_loop_thread(activate_connection_async)
        return future

    def stop_connection_async(self, connection: NM.ActiveConnection) -> Future:
        """Stops a VPN connection asynchronously.
        :param connection: connection to be stopped.
        :return: Future to know when the connection has been stopped.
        """
        callback, future = self.create_nmcli_callback(
            finish_method_name="deactivate_connection_finish"
        )

        def deactivate_connection_async():
            self._assert_running_on_main_loop_thread()
            self._nm_client.deactivate_connection_async(
                connection,
                None,
                callback,
                None
            )

        self._run_on_main_loop_thread(deactivate_connection_async)
        return future

    def remove_connection_async(
            self, connection: NM.RemoteConnection
    ) -> Future:
        """
        Removes the specified connection asynchronously.
        https://lazka.github.io/pgi-docs/#NM-1.0/classes/RemoteConnection.html#NM.RemoteConnection.delete_async
        :param connection: connection to be removed.
        :return: a Future to keep track of completion.
        """
        callback, future = self.create_nmcli_callback(
            finish_method_name="delete_finish"
        )

        def delete_async():
            self._assert_running_on_main_loop_thread()
            connection.delete_async(
                None,
                callback,
                None
            )

        self._run_on_main_loop_thread(delete_async)
        return future

    def get_active_connection(self, uuid: str) -> Optional[NM.ActiveConnection]:
        """
        Returns the specified active connection, if existing.
        :param uuid: UUID of the active connection.
        :return: the active connection if it was found. Otherwise, None.
        """
        active_connections = self._nm_client.get_active_connections()

        for connection in active_connections:
            if connection.get_uuid() == uuid:
                return connection

        return None

    def get_connection(self, uuid: str) -> Optional[NM.RemoteConnection]:
        """
        Returns the specified connection, if existing.
        :param uuid: UUID of the connection.
        :return: the connection if it was found. Otherwise, None.
        """
        return self._nm_client.get_connection_by_uuid(uuid)



"""
This module contains the backends implementing the supported OpenVPN protocols.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import os
from random import randrange
from typing import Optional, Union
from ipaddress import IPv4Address, IPv6Address
import socket
import uuid
import logging
from getpass import getuser
from concurrent.futures import Future
from dataclasses import dataclass

import gi

gi.require_version("NM", "1.0")  # noqa: required before importing NM module
# pylint: disable=wrong-import-position
from gi.repository import NM

from proton.vpn.connection import events, states
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.interfaces import Settings
from proton.vpn.backend.networkmanager.core import (LinuxNetworkManager, LocalAgentMixin)


logger = logging.getLogger(__name__)

FWMARK_ENV_VAR = "PROTON_VPN_FWMARK"
MIN_FWMARK_VALUE = 51821
MAX_FWMARK_VALUE = 2**32  # 32-bit integer


@dataclass
class Config:
    """Contains all specific details for networking configuration."""
    address: str
    address_prefix: int
    dns_ip: str
    allowed_ip: str
    dns_search: str = "~"
    dns_priority: int = -1500


@dataclass
class WireGuardConfig:
    """Contains networking configurations for both IPv4/6."""
    ipv4: Config
    ipv6: Config

    def get_dns_ip_for_protocol_version(self, ip_version: Union[IPv4Address, IPv6Address]):
        """Returns dns IP value based on IP version."""
        if ip_version == IPv4Address:
            return self.ipv4.dns_ip
        return self.ipv6.dns_ip

    def get_dns_search_for_protocol_version(self, ip_version: Union[IPv4Address, IPv6Address]):
        """Returns dns search value based on IP version."""
        if ip_version == IPv4Address:
            return self.ipv4.dns_search
        return self.ipv6.dns_search


wg_config = WireGuardConfig(
    ipv4=Config(
        address="10.2.0.2",
        address_prefix=32,
        dns_ip="10.2.0.1",
        allowed_ip="0.0.0.0/0",
    ),
    ipv6=Config(
        address="2a07:b944::2:2",
        address_prefix=128,
        dns_ip="2a07:b944::2:1",
        allowed_ip="::/0",
    )
)


def get_fwmark_from_env_var() -> Optional[int]:
    """
    Returns the fwmark from the env var or None if not available or not valid.
    """
    fwmark_str = os.getenv(FWMARK_ENV_VAR)

    if not fwmark_str:
        return None

    try:
        fwmark = int(fwmark_str)

        if fwmark not in range(MIN_FWMARK_VALUE, MAX_FWMARK_VALUE):
            raise ValueError("fwmark out of range")

        return fwmark
    except ValueError:
        logger.error(
            "The %s env var should contain an integer "
            "higher or equal than %s and lower than %s",
            FWMARK_ENV_VAR, MIN_FWMARK_VALUE, MAX_FWMARK_VALUE
        )

    return None


def get_random_fwmark() -> int:
    """Returns a random fwmark within the expected range."""
    # nosemgrep: gitlab.bandit.B311
    return randrange(MIN_FWMARK_VALUE, MAX_FWMARK_VALUE)  # nosec B311


class Wireguard(LinuxNetworkManager, LocalAgentMixin):
    """Creates a Wireguard connection."""
    SIGNAL_NAME: str = "state-changed"
    VIRTUAL_DEVICE_NAME: str = "proton0"
    protocol: str = "wireguard"
    ui_protocol: str = "WireGuard"
    connection: Optional[NM.SimpleConnection] = None
    FWMARK: int = get_fwmark_from_env_var() or get_random_fwmark()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        LocalAgentMixin.__init__(self)
        self._connection_settings = None

    def setup(self) -> Future:
        """Methods that creates and applies any necessary changes to the connection."""
        self._generate_connection()
        self._modify_connection()
        return self.nm_client.add_connection_async(self.connection)

    def _generate_connection(self):
        self._unique_id = str(uuid.uuid4())
        self._connection_settings = NM.SettingConnection.new()
        self.connection = NM.SimpleConnection.new()

    async def update_credentials(self, credentials):
        """Notifies the vpn server that the wireguard certificate needs a refresh."""
        await super().update_credentials(credentials)
        await self._start_local_agent_listener()

    @property
    def are_feature_updates_applied_when_active(self) -> bool:
        """
        Returns whether the connection features updates are applied on the fly
        while the connection is already active, without restarting the connection.
        """
        return True

    async def update_settings(self, settings: Settings):
        """Update features on the active agent connection."""
        await super().update_settings(settings)
        if self._agent_listener.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            await self._request_connection_features(settings.features)

    def _modify_connection(self):
        self._set_custom_connection_id()
        self._set_uuid()
        self._set_interface_name()
        self._set_connection_type()
        self._set_connection_user_owned()
        self.connection.add_setting(self._connection_settings)

        self._set_route()
        self._set_dns()
        self._set_wireguard_properties()

        self.connection.verify()

    def _set_custom_connection_id(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_ID, self._get_servername())

    def _set_uuid(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_UUID, self._unique_id)

    def _set_interface_name(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_INTERFACE_NAME, self.VIRTUAL_DEVICE_NAME
        )

    def _set_connection_type(self):
        self._connection_settings.set_property(
            NM.SETTING_CONNECTION_TYPE, NM.SETTING_WIREGUARD_SETTING_NAME
        )

    def _set_connection_user_owned(self):
        self._connection_settings.add_permission(
            NM.SETTING_USER_SETTING_NAME,
            getuser(),
            None
        )

    def _set_route(self):
        ipv4_config = NM.SettingIP4Config.new()
        ipv6_config = NM.SettingIP6Config.new()

        ipv4_config.set_property(NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP4_CONFIG_METHOD_MANUAL)
        ipv4_config.add_address(
            NM.IPAddress.new(socket.AF_INET, wg_config.ipv4.address, wg_config.ipv4.address_prefix)
        )

        if self.enable_ipv6_support:
            ipv6_config.set_property(
                NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP6_CONFIG_METHOD_MANUAL
            )
            ipv6_config.add_address(
                NM.IPAddress.new(
                    socket.AF_INET6, wg_config.ipv6.address, wg_config.ipv6.address_prefix
                )
            )
        else:
            ipv6_config.set_property(
                NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP6_CONFIG_METHOD_DISABLED
            )

        self.connection.add_setting(ipv4_config)
        self.connection.add_setting(ipv6_config)

    def _set_dns(self):
        ipv4_config = self.connection.get_setting_ip4_config()
        ipv6_config = self.connection.get_setting_ip6_config()

        self._configure_dns(nm_setting=ipv4_config, ip_version=IPv4Address)
        if self.enable_ipv6_support:
            self._configure_dns(nm_setting=ipv6_config, ip_version=IPv6Address)

        self.connection.add_setting(ipv4_config)
        self.connection.add_setting(ipv6_config)

    def _configure_dns(
        self,
        nm_setting: Union[NM.SettingIP4Config, NM.SettingIP6Config],
        ip_version: Union[IPv4Address, IPv6Address],
        dns_priority: int = -1500,
    ):
        """Sets DNS."""
        if ip_version not in [IPv4Address, IPv6Address]:
            raise ValueError(f"Unknown IP version: {ip_version}")

        nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, dns_priority)
        nm_setting.set_property(NM.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

        # pylint: disable=duplicate-code
        custom_dns_ips = self._settings.custom_dns\
            .get_enabled_dns_list_based_on_ip_version(ip_version)
        ip_addresses = [dns.exploded for dns in custom_dns_ips]

        # If custom DNS is disabled or there are no IP addresses then
        # we need to set anyway the DNS because WG does not handle it automatically
        # like OpenVPN does.
        if self._settings.custom_dns.enabled and ip_addresses:
            nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS, ip_addresses)
        else:
            nm_setting.add_dns(wg_config.get_dns_ip_for_protocol_version(ip_version))
            nm_setting.add_dns_search(wg_config.get_dns_search_for_protocol_version(ip_version))

    def _set_wireguard_properties(self):
        peer = NM.WireGuardPeer.new()
        wireguard_config = NM.SettingWireGuard.new()

        peer.append_allowed_ip(wg_config.ipv4.allowed_ip, False)
        peer.set_endpoint(
            f"{self._vpnserver.server_ip}:{self._vpnserver.wireguard_ports.udp[0]}",
            False
        )

        peer.set_public_key(self._vpnserver.x25519pk, False)

        if self.enable_ipv6_support:
            peer.append_allowed_ip(wg_config.ipv6.allowed_ip, False)

        # Seal the NM.WireGuardPeer instance. Afterwards, it is a bug to call all functions that
        # modify the instance (except ref/unref). A sealed instance cannot be unsealed again,
        # but you can create an unsealed copy with NM.WireGuardPeer.new_clone().
        # https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/WireGuardPeer.html#NM.WireGuardPeer.seal
        peer.seal()

        # Ensures that the configurations are valid
        # https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/WireGuardPeer.html#NM.WireGuardPeer.is_valid
        peer.is_valid(True, True)
        wireguard_config.append_peer(peer)

        wireguard_config.set_property(
            NM.SETTING_WIREGUARD_PRIVATE_KEY,
            self._vpncredentials.pubkey_credentials.wg_private_key
        )

        wireguard_config.set_property(
            NM.SETTING_WIREGUARD_FWMARK, self.FWMARK
        )

        self.connection.add_setting(wireguard_config)

    # pylint: disable=arguments-renamed
    def _on_state_changed(
            self, _: NM.ActiveConnection, state: int, reason: int
    ):
        """
            When the connection state changes, NM emits a signal with the state and
            reason for the change. This callback will receive these updates
            and translate for them accordingly for the state machine,
            as the state machine is backend agnostic.

            :param state: connection state update
            :type state: int
            :param reason: the reason for the state update
            :type reason: int
        """
        state = NM.ActiveConnectionState(state)
        reason = NM.ActiveConnectionStateReason(reason)

        logger.debug(
            "Wireguard connection state changed: state=%s, reason=%s",
            state.value_name, reason.value_name
        )

        if state is NM.ActiveConnectionState.ACTIVATED:
            self._async_start_local_agent_listener()
        elif state == NM.ActiveConnectionState.DEACTIVATED:
            self._async_stop_local_agent_listener()
            self._notify_subscribers_threadsafe(
                events.Disconnected(EventContext(connection=self))
            )
        else:
            logger.debug("Ignoring VPN state change: %s", state.value_name)

    def _initialize_persisted_connection(
            self, connection_id: str
    ) -> states.State:
        """Implemented in wireguard so we can start local agent listener."""
        state = super()._initialize_persisted_connection(connection_id)

        if isinstance(state, states.Connected):
            self._async_start_local_agent_listener()
        return state

    @classmethod
    def _get_priority(cls):
        return 1

    @classmethod
    def _validate(cls):
        # FIX ME: This should do a validation to ensure that NM can be used
        return True



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from proton.vpn import logging
from proton.vpn.session.client_config import ClientConfigFetcher, ClientConfig
from proton.vpn.session.credentials import VPNPubkeyCredentials
from proton.vpn.session.dataclasses import (
    VPNCertificate, VPNSessions, VPNSettings,
    VPNLocation
)
from proton.vpn.session.servers.server_list_fetcher import ServerListFetcher
from proton.vpn.session.servers.server_list_fetcher import EndpointVersion
from proton.vpn.session.servers.logicals import ServerList
from proton.vpn.session.utils import rest_api_request
from proton.vpn.session.feature_flags_fetcher import FeatureFlagsFetcher, FeatureFlags

from proton.vpn.core.settings.features import Features

if TYPE_CHECKING:
    from proton.vpn.session import VPNSession

logger = logging.getLogger(__name__)

# These are the api keys for the certificate features.
API_NETSHIELD = "NetShieldLevel"
API_VPN_ACCELERATOR = "SplitTCP"
API_MODERATE_NAT = "RandomNAT"
API_PORT_FORWARDING = "PortForwarding"


class VPNSessionFetcher:
    """
    Fetches PROTON VPN user account information.
    """
    # Note that the API does not allow intervals shorter than 1 day.
    _CERT_DURATION_IN_MIN = VPNPubkeyCredentials.REFRESH_INTERVAL // 60

    def __init__(
            self, session: "VPNSession",
            server_list_fetcher: Optional[ServerListFetcher] = None,
            client_config_fetcher: Optional[ClientConfigFetcher] = None,
            features_fetcher: Optional[FeatureFlagsFetcher] = None,
    ):
        self._session = session
        self._server_list_fetcher = server_list_fetcher or ServerListFetcher(session)
        self._client_config_fetcher = client_config_fetcher or ClientConfigFetcher(session)
        self._feature_flags_fetcher = features_fetcher or FeatureFlagsFetcher(session)

    async def fetch_vpn_info(self) -> VPNSettings:
        """Fetches client VPN information."""
        return VPNSettings.from_dict(
            await rest_api_request(self._session, "/vpn/v2")
        )

    async def fetch_certificate(
        self, client_public_key,
        features: Optional[Features] = None
    ) -> VPNCertificate:
        """
        Fetches a certificated signed by the API server to authenticate against VPN servers.
        """
        json_req = {
            "ClientPublicKey": client_public_key,
            "Duration": f"{self._CERT_DURATION_IN_MIN} min"
        }
        if features:
            json_req["Features"] = VPNSessionFetcher._convert_features(features)

        return VPNCertificate.from_dict(
            await rest_api_request(
                self._session, "/vpn/v1/certificate", jsondata=json_req
            )
        )

    async def fetch_active_sessions(self) -> VPNSessions:
        """
        Fetches information about active VPN sessions.
        """
        return VPNSessions.from_dict(
            await rest_api_request(self._session, "/vpn/v1/sessions")
        )

    async def fetch_location(self) -> VPNLocation:
        """Fetches information about the physical location the VPN client is connected from."""
        return VPNLocation.from_dict(
            await rest_api_request(self._session, "/vpn/v1/location")
        )

    def load_server_list_from_cache(self) -> ServerList:
        """
        Loads the previously persisted server list.
        :returns: the loaded server lists.
        :raises ServerListDecodeError: if the server list could not be loaded.
        """
        return self._server_list_fetcher.load_from_cache()

    async def fetch_server_list(
            self, endpoint_version: EndpointVersion) -> ServerList:
        """Fetches the list of VPN servers."""
        return await self._server_list_fetcher.fetch(endpoint_version)

    async def update_server_loads(
            self, endpoint_version: EndpointVersion) -> ServerList:
        """Fetches new server loads and updates the current server list with them."""
        return await self._server_list_fetcher.update_loads(endpoint_version)

    def load_client_config_from_cache(self) -> ClientConfig:
        """
        Loads the previously persisted client configuration.
        :returns: the loaded client configuration.
        :raises ClientConfigDecodeError: if the client configuration could not be loaded.
        """
        return self._client_config_fetcher.load_from_cache()

    async def fetch_client_config(self) -> ClientConfig:
        """Fetches general client configuration to connect to VPN servers."""
        return await self._client_config_fetcher.fetch()

    def load_feature_flags_from_cache(self) -> FeatureFlags:
        """
        Loads the previously persisted client configuration.
        :returns: the loaded client configuration.
        :raises ClientConfigDecodeError: if the client configuration could not be loaded.
        """
        return self._feature_flags_fetcher.load_from_cache()

    async def fetch_feature_flags(self) -> FeatureFlags:
        """Fetches general client configuration to connect to VPN servers."""
        return await self._feature_flags_fetcher.fetch()

    def clear_cache(self):
        """Discards the cache, if existing."""
        self._server_list_fetcher.clear_cache()
        self._client_config_fetcher.clear_cache()
        self._feature_flags_fetcher.clear_cache()

    @staticmethod
    def _convert_features(features: Features):
        """
        This converts the settings features into a certificate request features
        dictionary.
        """
        result = {}

        if not features.moderate_nat:
            result[API_MODERATE_NAT] = False

        if not features.vpn_accelerator:
            result[API_VPN_ACCELERATOR] = False

        if features.port_forwarding:
            result[API_PORT_FORWARDING] = True

        if features.netshield != 0:
            result[API_NETSHIELD] = features.netshield

        return result



"""
The public interface and the functionality that's common to all supported
VPN connection backends is defined in this module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("proton-vpn-connection")
except PackageNotFoundError:
    __version__ = "development"


# pylint: disable=wrong-import-position
from .vpnconnection import VPNConnection
from .interfaces import (
    VPNServer, ProtocolPorts, VPNCredentials, VPNPubkeyCredentials,
    VPNUserPassCredentials, Settings
)

__all__ = [
    "VPNConnection", "VPNServer", "ProtocolPorts", "VPNCredentials",
    "VPNPubkeyCredentials", "VPNUserPassCredentials", "Settings"
]



"""
Proton VPN API.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import copy
from threading import Event
from typing import Optional

from proton.vpn import logging
from proton.vpn.core.connection import VPNConnector
from proton.vpn.core.refresher.scheduler import Scheduler
from proton.vpn.core.refresher.vpn_data_refresher import VPNDataRefresher
from proton.vpn.core.settings import Settings, SettingsPersistence
from proton.vpn.core.session_holder import SessionHolder, ClientTypeMetadata
from proton.vpn.session.dataclasses import LoginResult, BugReportForm
from proton.vpn.session.account import VPNAccount
from proton.vpn.session import FeatureFlags
from proton.vpn.core.usage import UsageReporting

from proton.session.api import Fido2Assertion

from proton.vpn.session.u2f_interaction import UserInteraction

logger = logging.getLogger(__name__)


class ProtonVPNAPI:  # pylint: disable=too-many-public-methods
    """Class exposing the Proton VPN facade."""
    def __init__(self, client_type_metadata: ClientTypeMetadata):
        self._session_holder = SessionHolder(
            client_type_metadata=client_type_metadata
        )
        self._settings_persistence = SettingsPersistence()
        self._vpn_connector = None
        self._usage_reporting = UsageReporting(
            client_type_metadata=client_type_metadata)
        self.refresher = VPNDataRefresher(
            self._session_holder, Scheduler()
        )
        self._split_tunneling_client = None

    async def get_vpn_connector(self) -> VPNConnector:
        """Returns an object that wraps around the raw VPN connection object.

        This will provide some additional helper methods
        related to VPN connections and VPN servers.
        """
        if self._vpn_connector:
            return self._vpn_connector

        self._vpn_connector = await VPNConnector.get(
            session_holder=self._session_holder,
            settings_persistence=self._settings_persistence,
            usage_reporting=self._usage_reporting,
        )
        self._vpn_connector.subscribe_to_certificate_updates(self.refresher)

        return self._vpn_connector

    async def load_settings(self) -> Settings:
        """
        Returns a copy of the settings saved to disk, or the defaults if they
        are not found. Be sure to call save_settings if you want to apply changes.
        """
        # Default to free user settings if the session is not loaded yet.
        # pylint: disable=duplicate-code
        user_tier = self._session_holder.user_tier or 0

        loop = asyncio.get_running_loop()
        settings = await loop.run_in_executor(
            None, self._settings_persistence.get,
            user_tier
        )
        self._usage_reporting.enabled = settings.anonymous_crash_reports

        # We have to return a copy of the settings to force the caller to
        # use the `save_settings` method to apply the changes.
        return copy.deepcopy(settings)

    async def save_settings(self, settings: Settings):
        """
        Saves the settings to disk.

        Certain actions might be triggered by the VPN connector. For example, the
        kill switch might also be enabled/disabled depending on the setting value.
        """
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._settings_persistence.save, settings)
        await self._vpn_connector.apply_settings(settings)
        self._usage_reporting.enabled = settings.anonymous_crash_reports

    async def login(self, username: str, password: str) -> LoginResult:
        """
        Logs the user in provided the right credentials.
        :param username: Proton account username.
        :param password: Proton account password.
        :return: The login result.
        """
        session = self._session_holder.get_session_for(username)

        result = await session.login(username, password)
        if result.success and not session.loaded:
            await session.fetch_session_data()

        return result

    async def submit_2fa_code(self, code: str) -> LoginResult:
        """
        Submits the 2-factor authentication code.
        :param code: 2FA code.
        :return: The login result.
        """
        session = self._session_holder.session
        result = await session.provide_2fa_code(code)

        if result.success and not session.loaded:
            await session.fetch_session_data()

        return result

    @property
    def is_fido2_lib_available(self) -> bool:
        """
        Returns whether we support U2F/FIDO2 security keys for 2FA on this platform.

        This is deprecated, use is_fido2_available instead.
        """
        logger.warning("is_fido2_lib_available is deprecated, use is_fido2_available instead")
        return bool(self._session_holder.session.fido2_lib_available)

    @property
    def supports_fido2(self) -> bool:
        """
        Returns if
        - We support U2F/FIDO2 security keys for 2FA on this platform.
        - The user has fido2 keys registered.

        This only returns True if both conditions are met and if the user
        is currently authenticating a session that requires 2FA.
        """
        lib_available = self.is_fido2_lib_available
        supports_fido2 = self._session_holder.session.supports_fido2
        return bool(lib_available and supports_fido2)

    async def generate_2fa_fido2_assertion(
            self,
            user_interaction: Optional[UserInteraction] = None,
            cancel_assertion: Optional[Event] = None
    ) -> Fido2Assertion:
        """
        Generates a 2FA assertion using a U2F/FIDO2 security key.

        :param user_interaction: object handling any required user interaction
            while generating the assertion.
        :param cancel_assertion: optional event that can be set to cancel the
        fido 2 assertion process.
        :returns: the generated FIDO 2 assertion.
        """
        return await self._session_holder.session.generate_2fa_fido2_assertion(
            user_interaction, cancel_assertion
        )

    async def submit_2fa_fido2(self, fido2_assertion: Fido2Assertion) -> LoginResult:
        """
        Submits the 2-factor authentication using a U2F/FIDO2 security key.
        :return: The login result.
        """
        session = self._session_holder.session
        result = await session.provide_2fa_fido2(fido2_assertion)

        if result.success and not session.loaded:
            await session.fetch_session_data()

        return result

    def is_user_logged_in(self) -> bool:
        """Returns True if a user is logged in and False otherwise."""
        return self._session_holder.session.logged_in

    @property
    def account_name(self) -> str:
        """Returns account name."""
        return self._session_holder.session.AccountName

    @property
    def account_data(self) -> VPNAccount:
        """
        Returns account data, which contains information such
        as (but not limited to):
         - Plan name/title
         - Max tier
         - Max connections
         - VPN Credentials
         - Location
        """
        return self._session_holder.session.vpn_account

    @property
    def user_tier(self) -> int:
        """
        Returns the Proton VPN tier.

        Current possible values are:
         * 0: Free
         * 2: Plus
         * 3: Proton employee

        Note: tier 1 is no longer in use.
        """
        return self.account_data.max_tier

    @property
    def vpn_session_loaded(self) -> bool:
        """Returns whether the VPN session data was already loaded or not."""
        return self._session_holder.session.loaded

    @property
    def server_list(self):
        """The last server list fetched from the REST API."""
        return self._session_holder.session.server_list

    @property
    def client_config(self):
        """The last client configuration fetched from the REST API."""
        return self._session_holder.session.client_config

    @property
    def feature_flags(self) -> FeatureFlags:
        """The last feature flags fetched from the REST API."""
        return self._session_holder.session.feature_flags

    async def submit_bug_report(self, bug_report: BugReportForm):
        """
        Submits the specified bug report to customer support.
        """
        return await self._session_holder.session.submit_bug_report(bug_report)

    async def logout(self):
        """
        Logs the current user out.
        :raises: VPNConnectionFoundAtLogout if the users is still connected to the VPN.
        """
        await self.refresher.disable()
        await self._session_holder.session.logout()
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(executor=None, func=self._settings_persistence.delete)
        vpn_connector = await self.get_vpn_connector()
        await vpn_connector.disconnect()

    @property
    def usage_reporting(self) -> UsageReporting:
        """Returns the usage reporting instance to send anonymous crash reports."""
        return self._usage_reporting

    @property
    def split_tunneling_available(self) -> bool:
        """Deprecated, use VPNConnector.is_split_tunneling_available."""
        logger.warning("Deprecated: use VPNConnector.is_split_tunneling_available instead")
        # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses  # pylint: disable=line-too-long  # noqa: E501
        return self._vpn_connector.is_split_tunneling_available



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum
from pathlib import Path
from typing import Optional, TYPE_CHECKING, Dict, Tuple
import re

from proton.utils.environment import VPNExecutionEnvironment

from proton.vpn.core.cache_handler import CacheHandler
from proton.vpn.session.exceptions import ServerListDecodeError
from proton.vpn.session.servers.types import ServerLoad
from proton.vpn.session.servers.logicals import ServerList, PersistenceKeys
from proton.vpn.session.utils import rest_api_request

if TYPE_CHECKING:
    from proton.vpn.session import VPNSession
    from proton.vpn.session.dataclasses import VPNLocation

try:
    # The import of the proton.vpn.lib module is delayed to avoid
    # importing it when the feature flag is disabled.
    from proton.vpn.lib import ServerStatus  # pylint: disable=C0415, E0401, E0611
    VPN_LIB_AVAILABLE = True
except ImportError:
    VPN_LIB_AVAILABLE = False

NETZONE_HEADER = "X-PM-netzone"
MODIFIED_SINCE_HEADER = "If-Modified-Since"
LAST_MODIFIED_HEADER = "Last-Modified"
NOT_MODIFIED_STATUS = 304


class EndpointVersion(Enum):
    """Used to choose between v1 endpoint or v2"""
    V1 = 1
    V2 = 2


class MixinEndpointV1:  # pylint: disable=R0903
    """
    Mixin class for the v1 endpoints of the Proton VPN REST API.
    """
    LOGICALS = "/vpn/v1/logicals?SecureCoreFilter=all"
    LOADS = "/vpn/v1/loads"

    async def _v1_fetch_logicals(self) -> Tuple[Dict, str]:
        return await self._request_logicals(MixinEndpointV1.LOGICALS)

    async def _v1_update_loads(self) -> ServerList:
        """
        Fetches the server loads from the REST API and
        updates the current server list with them."""
        if not self._server_list:
            raise RuntimeError(
                "Server loads can only be updated after fetching the the full server list."
            )

        response = await rest_api_request(
            self._session,
            self.LOADS,
            additional_headers=self._build_additional_headers(),
        )

        server_loads = [ServerLoad(data) for data in response["LogicalServers"]]
        self._server_list.update(server_loads)
        self._cache_file.save(self._server_list.to_dict())

        return self._server_list


class MixinEndpointV2:  # pylint: disable=R0903
    """
    Mixin class for the v2 endpoints of the Proton VPN REST API.
    """
    LOGICALS = "/vpn/v2/logicals?SecureCoreFilter=all"
    STATUS = "/vpn/v2/status/{token}/binary"

    def _convert_load(self, load: Dict) -> Dict:

        # Going forward we want to properly integrate, IsEnabled, IsVisible and
        # IsAutoconnectable see VPNLINUX-1502.
        score = load["Score"] + (0 if load["IsAutoconnectable"] else 1000)
        return {
            "Load": load["Load"],
            "Score": score,
            "Status": load["IsEnabled"] and load["IsVisible"],
        }

    async def _v2_fetch_logicals(self, location: Optional["VPNLocation"]) -> Tuple[Dict, str]:

        logicals, last_modified_time =\
            await self._request_logicals(MixinEndpointV2.LOGICALS)

        self._server_status = ServerStatus(
            logicals,
            user_location={
                "Latitude": location.Lat,
                "Longitude": location.Long
            },
            user_country=location.Country
        )

        status = await self._request_status(
            MixinEndpointV2.STATUS.format(token=logicals["StatusID"])
        )

        loads = self._server_status.compute_loads(status)

        # Splice the loads into the servers
        servers = logicals["LogicalServers"]
        if len(loads) != len(servers):
            raise RuntimeError(
                "Loads computation produced a different number of servers "
                "than the logicals list. This is unexpected."
            )

        for server, load in zip(servers, loads):
            server.update(self._convert_load(load))

        return logicals, last_modified_time

    async def _v2_update_loads(self, status) -> ServerList:
        """
        Fetches the server loads from the REST API and
        updates the current server list with them."""
        if not self._server_list:
            raise RuntimeError(
                "Server loads can only be updated after fetching the the full server list."
            )

        binary_status = await self._request_status(
            MixinEndpointV2.STATUS.format(token=status)
        )

        loads = self._server_status.compute_loads(binary_status)
        server_loads = [ServerLoad(self._convert_load(data)) for data in loads]

        self._server_list.update(server_loads)
        self._cache_file.save(self._server_list.to_dict())

        return self._server_list


class ServerListFetcher(MixinEndpointV1, MixinEndpointV2):
    """Fetches the server list either from disk or from the REST API."""

    CACHE_PATH = Path(VPNExecutionEnvironment().path_cache) / "serverlist.json"

    """Fetches and caches the list of VPN servers from the REST API."""
    def __init__(
            self,
            session: "VPNSession",
            server_list: Optional[ServerList] = None,
            cache_file: Optional[CacheHandler] = None
    ):
        self._session = session
        self._server_list = server_list
        self._cache_file = cache_file or CacheHandler(self.CACHE_PATH)

    def clear_cache(self):
        """Discards the cache, if existing."""
        self._server_list = None
        self._cache_file.remove()

    async def _request_logicals(self, endpoint: str) -> Tuple[Dict, str]:
        raw_response = await rest_api_request(
            self._session,
            endpoint,
            additional_headers=self._build_additional_headers(),
            return_raw=True
        )

        if raw_response.status_code == NOT_MODIFIED_STATUS:
            response = self._server_list.to_dict()
        else:
            response = raw_response.json

        # The last modified time
        last_modified_time = raw_response.find_first_header(
            LAST_MODIFIED_HEADER, ServerList.get_epoch_time()
        )

        return response, last_modified_time

    async def _request_status(self, endpoint):
        status_response = await rest_api_request(
            self._session,
            endpoint,
            additional_headers=self._build_additional_headers(),
            return_raw=True
        )

        return status_response.data

    def _v2_validate_location(self) -> Optional["VPNLocation"]:
        location = self._session.vpn_account.location
        if (location.Lat is None or location.Long is None):
            return None

        return location

    async def fetch(self, endpoint_version: EndpointVersion) -> ServerList:
        """Fetches the list of VPN servers. Warning: this is a heavy request."""

        location = self._v2_validate_location()
        use_v2 = endpoint_version == EndpointVersion.V2 \
            and VPN_LIB_AVAILABLE\
            and location is not None

        if use_v2:
            response, last_modified_time = await self._v2_fetch_logicals(location)
        else:
            response, last_modified_time = await self._v1_fetch_logicals()

        Keys = PersistenceKeys
        entries_to_update = {
            Keys.USER_TIER.value: self._session.vpn_account.max_tier,
            Keys.LAST_MODIFIED_TIME.value: last_modified_time,
            Keys.EXPIRATION_TIME.value: ServerList.get_expiration_time(),
            Keys.LOADS_EXPIRATION_TIME.value:
                ServerList.get_loads_expiration_time(),
        }

        response.update(entries_to_update)

        self._cache_file.save(response)

        self._server_list = ServerList.from_dict(response)

        return self._server_list

    async def update_loads(self, endpoint_version: EndpointVersion) -> ServerList:
        """
        Queries the REST API for the latest server loads and updates
        the current server list with them, return the updated server list.
        """
        status_token = self._server_list.status_token
        if status_token and (endpoint_version == EndpointVersion.V2):
            server_list = await self._v2_update_loads(status_token)
        else:
            server_list = await self._v1_update_loads()

        return server_list

    def load_from_cache(self) -> ServerList:
        """
        Loads and returns the server list that was last persisted to the cache.

        :returns: the server list loaded from cache.
        :raises ServerListDecodeError: if the cache is not found or if the
            data stored in the cache is not valid.
        """
        cache = self._cache_file.load()

        if not cache:
            raise ServerListDecodeError("Cached server list was not found")

        self._server_list = ServerList.from_dict(cache)
        return self._server_list

    def _build_additional_headers(self) -> Dict[str, str]:
        headers = {}
        headers[NETZONE_HEADER] = self._build_header_netzone()
        headers[MODIFIED_SINCE_HEADER] = self._extract_modified_since_header()

        return headers

    def _build_header_netzone(self) -> str:
        truncated_ip_address = truncate_ip_address(
            self._session.vpn_account.location.IP
        )
        return truncated_ip_address

    def _extract_modified_since_header(self) -> str:
        return self._server_list.last_modified_time \
            if self._server_list \
            else ServerList.get_epoch_time()


def truncate_ip_address(ip_address: str) -> str:
    """
    Truncates the last octet of the specified IP address and returns it.
    """
    match = re.match("(\\d+\\.\\d+\\.\\d+)\\.\\d+", ip_address)
    if not match:
        raise ValueError(f"Invalid IPv4 address: {ip_address}")

    # Replace the last byte with a zero to truncate the IP.
    truncated_ip = f"{match[1]}.0"

    return truncated_ip



"""
This module contains all configurations needed to create
connection Kill Switches for NM.Client.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code

from dataclasses import dataclass, field
import uuid

import gi
gi.require_version("NM", "1.0")
from gi.repository import NM, GLib  # pylint: disable=C0413 # noqa: E402


DEFAULT_METRIC = -1


@dataclass
class KillSwitchGeneralConfig:  # pylint: disable=missing-class-docstring
    human_readable_id: str
    interface_name: str


@dataclass
class KillSwitchIPConfig:  # pylint: disable=missing-class-docstring
    addresses: list
    dns: list
    dns_priority: str
    ignore_auto_dns: bool
    route_metric: str
    gateway: str = None
    routes: list = field(default_factory=list)


class KillSwitchConnection:  # pylint: disable=too-few-public-methods
    """Connection that is used to configure different types of Kill Switch
    connection. Are easily configured with the help of `KillSwitchGeneralConfig`
    and `KillSwitchIPConfig`.
    """
    def __init__(
        self,
        general_settings: KillSwitchGeneralConfig,
        ipv6_settings: KillSwitchIPConfig,
        ipv4_settings: KillSwitchIPConfig
    ):
        self._connection_profile = None
        self._general_settings = general_settings
        self._ipv6_settings = ipv6_settings
        self._ipv4_settings = ipv4_settings

    @property
    def connection(self) -> NM.Connection:
        """Lazy return connection object"""
        if self._connection_profile is None:
            self._create_connection_profile()

        return self._connection_profile

    def _create_connection_profile(self):
        self._connection_profile = NM.SimpleConnection.new()

        s_con = NM.SettingConnection.new()
        s_con.set_property(NM.SETTING_CONNECTION_ID, self._general_settings.human_readable_id)
        s_con.set_property(
            NM.SETTING_CONNECTION_INTERFACE_NAME,
            self._general_settings.interface_name
        )
        s_con.set_property(NM.SETTING_CONNECTION_UUID, str(uuid.uuid4()))
        s_con.set_property(NM.SETTING_CONNECTION_TYPE, NM.SETTING_DUMMY_SETTING_NAME)

        s_dummy = NM.SettingDummy.new()

        s_ipv4 = self._generate_ipv4_settings()
        s_ipv6 = self._generate_ipv6_settings()

        self._connection_profile.add_setting(s_con)
        self._connection_profile.add_setting(s_ipv4)
        self._connection_profile.add_setting(s_ipv6)
        self._connection_profile.add_setting(s_dummy)

        # Ensures the properties get correct values
        # https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/Connection.html#NM.Connection.verify
        if not s_con.verify():
            raise RuntimeError("Connection has invalid properties")

    def _generate_ipv4_settings(self):
        """
        For documentation see:
        https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/SettingIPConfig.html#NM.SettingIPConfig
        https://lazka.github.io/pgi-docs/NM-1.0/classes/IPAddress.html#NM.IPAddress
        https://lazka.github.io/pgi-docs/NM-1.0/classes/IPRoute.html#NM.IPRoute.new
        """
        s_ip4 = NM.SettingIP4Config.new()

        if self._ipv4_settings is None:
            s_ip4.set_property(NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP4_CONFIG_METHOD_DISABLED)
            return s_ip4

        # inform NM that the IP configuration is manual
        s_ip4.set_property(NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP4_CONFIG_METHOD_MANUAL)

        # Add addresses
        for address in self._ipv4_settings.addresses:
            ipv4, prefix = str(address).split("/")
            s_ip4.add_address(
                NM.IPAddress.new(GLib.SYSDEF_AF_INET, ipv4, int(prefix))
            )

        # Add DNS
        for dns in self._ipv4_settings.dns:
            s_ip4.add_dns(dns)

        # Add routes
        for route in self._ipv4_settings.routes:
            ipv4, prefix = str(route).split("/")
            s_ip4.add_route(
                NM.IPRoute.new(
                    family=GLib.SYSDEF_AF_INET, dest=ipv4, prefix=int(prefix),
                    next_hop=None, metric=DEFAULT_METRIC
                )
            )

        # Rest of the configs
        s_ip4.props.dns_priority = self._ipv4_settings.dns_priority
        s_ip4.props.route_metric = self._ipv4_settings.route_metric
        s_ip4.props.ignore_auto_dns = self._ipv4_settings.ignore_auto_dns

        if self._ipv4_settings.gateway:
            s_ip4.props.gateway = self._ipv4_settings.gateway

        return s_ip4

    def _generate_ipv6_settings(self):
        """
        For documentation see:
        https://lazka.github.io/pgi-docs/index.html#NM-1.0/classes/SettingIPConfig.html#NM.SettingIPConfig
        https://lazka.github.io/pgi-docs/NM-1.0/classes/IPAddress.html#NM.IPAddress
        https://lazka.github.io/pgi-docs/NM-1.0/classes/IPRoute.html#NM.IPRoute.new
        """
        s_ip6 = NM.SettingIP6Config.new()

        if self._ipv6_settings is None:
            s_ip6.set_property(NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP6_CONFIG_METHOD_DISABLED)
            return s_ip6

        # inform NM that the IP configuration is manual
        s_ip6.set_property(NM.SETTING_IP_CONFIG_METHOD, NM.SETTING_IP6_CONFIG_METHOD_MANUAL)

        # Add addresses
        for address in self._ipv6_settings.addresses:
            ip, prefix = str(address).split("/")  # pylint: disable=invalid-name
            s_ip6.add_address(
                NM.IPAddress.new(GLib.SYSDEF_AF_INET6, ip, int(prefix))
            )

        # Add DNS
        for dns in self._ipv6_settings.dns:
            s_ip6.add_dns(dns)

        # Rest of the configs
        s_ip6.props.dns_priority = self._ipv6_settings.dns_priority
        s_ip6.props.route_metric = self._ipv6_settings.route_metric
        s_ip6.props.ignore_auto_dns = self._ipv6_settings.ignore_auto_dns

        if self._ipv6_settings.gateway:
            s_ip6.props.gateway = self._ipv6_settings.gateway

        return s_ip6



"""
This module manages the Proton VPN general settings.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union
from ipaddress import ip_address, IPv4Address, IPv6Address

from proton.vpn import logging

logger = logging.getLogger(__name__)


@dataclass
class CustomDNSEntry:
    """Custom DNS IP object."""
    ip: Union[IPv4Address, IPv6Address]  # pylint: disable=invalid-name
    enabled: bool = True

    @staticmethod
    def from_dict(data: dict[str, Union[bool, str]]) -> CustomDNSEntry:
        """Creates and returns `CustomDNSEntry` from the provided dict."""
        try:
            ip = data["ip"]  # pylint: disable=invalid-name
        except KeyError as excp:
            raise ValueError("Missing 'ip' in custom DNS entry") from excp

        try:
            converted_ip = ip_address(ip)
        except ValueError as excp:
            raise ValueError("Invalid custom DNS IP") from excp

        return CustomDNSEntry(
            ip=converted_ip,
            enabled=bool(data.get("enabled", True))
        )

    def convert_ip_to_short_format(self) -> str:
        """Converts long format IP to short format IP.

        Mainly for IPv6 addresses.
        """
        return self.ip.compressed

    @staticmethod
    def new_from_string(new_dns_ip: str, enabled: bool = True) -> CustomDNSEntry:
        """Returns a new CustomDNSEntry from a string IP.

        This is an alternative way to instantiate this class, allowing the user to
        pass only the string IP, which internally will validate and convert it to
        and IPv4Address/IPv6Address object.
        """
        try:
            converted_ip = ip_address(new_dns_ip)
        except ValueError as excp:
            raise ValueError("Invalid custom DNS IP") from excp

        return CustomDNSEntry(ip=converted_ip, enabled=enabled)

    def to_dict(self) -> dict[str, Union[str, bool]]:
        """Converts the class to dict."""
        return {
            "ip": self.ip.compressed,
            "enabled": self.enabled
        }


@dataclass
class CustomDNS:
    """Contains all settings related to custom DNS."""
    enabled: bool = False
    ip_list: list[CustomDNSEntry] = field(default_factory=list)  # type: ignore

    @staticmethod
    def from_dict(data: dict[str, Union[bool, str, list]]) -> CustomDNS:
        """Creates and returns `CustomDNS` from the provided dict."""
        default = CustomDNS.default()
        loaded_ip_list: dict[str, Union[bool, str]] = data.get("ip_list", default.ip_list)
        ip_list = []

        for dns_entry_dict in loaded_ip_list:
            try:
                dns_ip = CustomDNSEntry.from_dict(dns_entry_dict)
            except ValueError as excp:
                logger.warning(
                    msg=f"Invalid custom DNS entry: {dns_entry_dict} : {excp}")
            else:
                ip_list.append(dns_ip)

        return CustomDNS(
            enabled=data.get("enabled", default.enabled),
            ip_list=ip_list
        )

    @staticmethod
    def default() -> CustomDNS:  # pylint: disable=unused-argument
        """Creates and returns `CustomDNS` from default configurations."""
        return CustomDNS()

    def get_enabled_ipv4_ips(self) -> list[IPv4Address]:
        """Returns a list of IPv4 custom DNSs that are enabled."""
        return self.get_enabled_dns_list_based_on_ip_version(IPv4Address)

    def get_enabled_ipv6_ips(self) -> list[IPv6Address]:
        """Returns a list of IPv6 custom DNSs that are enabled."""
        return self.get_enabled_dns_list_based_on_ip_version(IPv6Address)

    def get_enabled_dns_list_based_on_ip_version(
            self, version: Union[IPv4Address, IPv6Address]
    ) -> list[Union[str, None]]:
        """Returns a list of IPs based on provided IP version."""
        dns_list = []
        for dns in self.ip_list:
            if isinstance(dns.ip, version) and dns.enabled:
                dns_list.append(dns.ip)

        return dns_list

    def to_dict(self) -> dict[str, Union[bool, list[dict[str, Union[str, bool]]]]]:
        """Converts the class to dict."""
        return {
            "enabled": self.enabled,
            "ip_list": [ip.to_dict() for ip in self.ip_list]
        }



"""
OpenVPN protocol implementations.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from concurrent.futures import Future
import secrets
from getpass import getuser
from ipaddress import IPv4Address, IPv6Address
from typing import Union
import logging

from jinja2 import Environment, BaseLoader

from gi.repository import NM
import gi
from proton.vpn.backend.networkmanager.core import (LinuxNetworkManager, LocalAgentMixin)
from proton.vpn.connection.vpnconfiguration import VPNConfiguration
from proton.vpn.connection.constants import CA_CERT
from proton.vpn.connection import events, states
from proton.vpn.connection.events import EventContext
from proton.vpn.connection.interfaces import Settings

gi.require_version("NM", "1.0")

logger = logging.getLogger(__name__)

SECRET_PASSWORD_FIELD = "password"    # pylint: disable=line-too-long # noqa: E501 # nosec B105 # nosemgrep: generic.secrets.gitleaks.hashicorp-tf-password.hashicorp-tf-password
SECRET_CERT_PASS_FIELD = "cert-pass"  # pylint: disable=line-too-long # noqa: E501 # nosec B105 # nosemgrep: generic.secrets.gitleaks.hashicorp-tf-password.hashicorp-tf-password

PASSPHRASE_SECRET_LENGTH = 16
OPENVPN_V2_TEMPLATE = """
# ==============================================================================
# Copyright (c) 2016-2020 Proton Technologies AG (Switzerland)
# Email: contact@protonvpn.com
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR # OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
# ==============================================================================

{%- if enable_ipv6_support %}
push-peer-info
setenv UV_IPV6 1
{%- endif %}

client
dev tun
proto {{ openvpn_protocol|lower }}

{% for ip in serverlist %}
{%- for port in openvpn_ports -%}
remote {{ ip }} {{ port }}
{% endfor %}
{% endfor -%}

remote-random
resolv-retry infinite
nobind
cipher AES-256-GCM
verb 3

tun-mtu 1500
mssfix 0
persist-key
persist-tun

reneg-sec 0

remote-cert-tls server


<ca>
{{ca_certificate}}
</ca>

<tls-crypt>
-----BEGIN OpenVPN Static key V1-----
6acef03f62675b4b1bbd03e53b187727
423cea742242106cb2916a8a4c829756
3d22c7e5cef430b1103c6f66eb1fc5b3
75a672f158e2e2e936c3faa48b035a6d
e17beaac23b5f03b10b868d53d03521d
8ba115059da777a60cbfd7b2c9c57472
78a15b8f6e68a3ef7fd583ec9f398c8b
d4735dab40cbd1e3c62a822e97489186
c30a0b48c7c38ea32ceb056d3fa5a710
e10ccc7a0ddb363b08c3d2777a3395e1
0c0b6080f56309192ab5aacd4b45f55d
a61fc77af39bd81a19218a79762c3386
2df55785075f37d8c71dc8a42097ee43
344739a0dd48d03025b0450cf1fb5e8c
aeb893d9a96d1f15519bb3c4dcb40ee3
16672ea16c012664f8a9f11255518deb
-----END OpenVPN Static key V1-----
</tls-crypt>

<cert>
{{cert}}
</cert>
<key>
{{priv_key}}
</key>
"""


class OVPNConfig(VPNConfiguration):
    """OpenVPN-specific configuration."""
    PROTOCOL = None
    EXTENSION = ".ovpn"

    def __init__(self, private_key_passphrase, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._private_key_passphrase = private_key_passphrase

    def generate(self) -> str:
        """Method that generates a vpn config file.

        Returns:
            string: configuration file
        """
        openvpn_ports = self._vpnserver.openvpn_ports
        ports = openvpn_ports.tcp if "tcp" == self.PROTOCOL else openvpn_ports.udp

        enable_ipv6_support = self._vpnserver.has_ipv6_support and self._settings.ipv6

        j2_values = {
            "enable_ipv6_support": enable_ipv6_support,
            "openvpn_protocol": self.PROTOCOL,
            "serverlist": [self._vpnserver.server_ip],
            "openvpn_ports": ports,
            "ca_certificate": CA_CERT,
        }

        j2_values["cert"] =\
            self._vpncredentials.pubkey_credentials.certificate_pem
        password = self._private_key_passphrase.encode()
        j2_values["priv_key"] = self._vpncredentials.pubkey_credentials.\
            get_ed25519_sk_pem(password=password)

        template =\
            (Environment(loader=BaseLoader, autoescape=True)  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.flask.security.xss.audit.direct-use-of-jinja2.direct-use-of-jinja2
                .from_string(OPENVPN_V2_TEMPLATE))

        return template.render(j2_values)


class OpenVPNTCPConfig(OVPNConfig):
    """Configuration for OpenVPN using TCP."""
    PROTOCOL = "tcp"


class OpenVPNUDPConfig(OVPNConfig):
    """Configuration for OpenVPN using UDP."""
    PROTOCOL = "udp"


PROTOCOLS = {
    "openvpn-tcp": OpenVPNTCPConfig,
    "openvpn-udp": OpenVPNUDPConfig,
}


class OpenVPN(LinuxNetworkManager, LocalAgentMixin):
    """Base class for the backends implementing the OpenVPN protocols."""
    VIRTUAL_DEVICE_NAME = "proton0"
    PLUGIN_NAME = "openvpn"
    connection = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        LocalAgentMixin.__init__(self)
        self._vpn_settings = None
        self._connection_settings = None

    def setup(self) -> Future:
        """Methods that creates and applies any necessary changes to the connection."""
        passphrase = self._make_passphrase()
        self._generate_connection(passphrase)
        self._modify_connection(passphrase)
        return self.nm_client.add_connection_async(self.connection)

    def _make_passphrase(self):
        return secrets.token_hex(PASSPHRASE_SECRET_LENGTH)

    def _generate_connection(self, private_key_passphrase):
        vpnconfig = PROTOCOLS[self.protocol](
            private_key_passphrase,
            self._vpnserver, self._vpncredentials, self._settings
        )

        self.connection = self._import_vpn_config(self.PLUGIN_NAME, vpnconfig)

        self._unique_id = self.connection.get_uuid()
        self._vpn_settings = self.connection.get_setting_vpn()
        self._connection_settings = self.connection.get_setting_connection()

    def _modify_connection(self, private_key_passphrase):
        """Configure imported vpn connection.
        """
        self._set_custom_connection_id()
        self._set_connection_user_owned()
        self._set_server_certificate_check()
        self._set_dns()
        self._set_vpn_cert_credentials(private_key_passphrase)

        self.connection.add_setting(self._connection_settings)

    async def update_credentials(self, credentials):
        await super().update_credentials(credentials)
        await self._start_local_agent_listener()

    def _set_custom_connection_id(self):
        self._connection_settings.set_property(NM.SETTING_CONNECTION_ID, self._get_servername())

    def _set_connection_user_owned(self):
        # returns NM.SettingConnection
        # https://lazka.github.io/pgi-docs/NM-1.0/classes/SettingConnection.html#NM.SettingConnection

        self._connection_settings.add_permission(
            NM.SETTING_USER_SETTING_NAME,
            getuser(),
            None
        )

    def _set_server_certificate_check(self):
        appened_domain = "name:" + self._vpnserver.domain
        self._vpn_settings.add_data_item(
            "verify-x509-name", appened_domain
        )

    def _set_dns(self):
        """Apply dns configurations to ProtonVPN connection."""
        # pylint: disable=duplicate-code
        ipv4_config = self.connection.get_setting_ip4_config()
        ipv6_config = self.connection.get_setting_ip6_config()

        self._configure_dns(nm_setting=ipv4_config, ip_version=IPv4Address)
        if self.enable_ipv6_support:
            self._configure_dns(nm_setting=ipv6_config, ip_version=IPv6Address)

        self.connection.add_setting(ipv4_config)
        self.connection.add_setting(ipv6_config)

    def _configure_dns(
        self,
        nm_setting: Union[NM.SettingIP4Config, NM.SettingIP6Config],
        ip_version: Union[IPv4Address, IPv6Address],
        dns_priority: int = -1500,
    ):
        """Sets DNS."""
        if ip_version not in [IPv4Address, IPv6Address]:
            raise ValueError(f"Unknown IP version: {ip_version}")

        nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, dns_priority)

        custom_dns_ips = self._settings.custom_dns\
            .get_enabled_dns_list_based_on_ip_version(ip_version)
        ip_addresses = [dns.exploded for dns in custom_dns_ips]

        # OpenVPN sets DNS automatically if nothing is passed.
        if self._settings.custom_dns.enabled and ip_addresses:
            nm_setting.set_property(NM.SETTING_IP_CONFIG_DNS, ip_addresses)
            nm_setting.set_property(NM.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

    def _set_vpn_cert_credentials(self, private_key_passphrase: str):
        """
        Add passphrase to decrypt the private key.
        """
        self._vpn_settings.add_secret(SECRET_CERT_PASS_FIELD,
                                      private_key_passphrase)

    def _initialize_persisted_connection(self, connection_id: str) -> states.State:
        state = super()._initialize_persisted_connection(connection_id)

        if isinstance(state, states.Connected):
            self._async_start_local_agent_listener()
        return state

    # pylint: disable=unused-argument
    def _on_state_changed(
            self, vpn_connection: NM.VpnConnection, state: int, reason: int
    ):
        """
            When the vpn state changes, NM emits a signal with the state and
            reason for the change. This callback will receive these updates
            and translate for them accordingly for the state machine,
            as the state machine is backend agnostic.

            Note this method is called from the thread running the GLib main loop.
            Interactions between code in this method and the asyncio loop must
            be done in a thread-safe manner.

            :param state: connection state update
            :type state: int
            :param reason: the reason for the state update
            :type reason: int
        """
        try:
            state = NM.VpnConnectionState(state)
        except ValueError:
            logger.warning("Unexpected VPN connection state: %s", state)
            state = NM.VpnConnectionState.UNKNOWN

        try:
            reason = NM.VpnConnectionStateReason(reason)
        except ValueError:
            logger.warning("Unexpected VPN connection state reason: %s", reason)
            reason = NM.VpnConnectionStateReason.UNKNOWN

        logger.debug(
            "VPN connection state changed: state=%s, reason=%s",
            state.value_name, reason.value_name
        )

        def start_local_agent():
            self._async_start_local_agent_listener()

        def stop_local_agent():
            self._async_stop_local_agent_listener()

        if state is NM.VpnConnectionState.ACTIVATED:
            start_local_agent()
            self._notify_subscribers_threadsafe(
                events.Connected(EventContext(connection=self))
            )
        elif state is NM.VpnConnectionState.FAILED:
            if reason in [
                NM.VpnConnectionStateReason.CONNECT_TIMEOUT,
                NM.VpnConnectionStateReason.SERVICE_START_TIMEOUT
            ]:
                self._notify_subscribers_threadsafe(
                    events.Timeout(EventContext(connection=self, reason=reason))
                )
            elif reason in [
                NM.VpnConnectionStateReason.NO_SECRETS,
                NM.VpnConnectionStateReason.LOGIN_FAILED
            ]:
                # NO_SECRETS is passed when the user cancels the NM popup
                # to introduce the OpenVPN password. If we switch auth to
                # certificates, we should treat NO_SECRETS as an
                # UnexpectedDisconnection event.
                self._notify_subscribers_threadsafe(
                    events.AuthDenied(EventContext(connection=self, reason=reason))
                )
            else:
                # reason in [
                #     NM.VpnConnectionStateReason.UNKNOWN,
                #     NM.VpnConnectionStateReason.NONE,
                #     NM.VpnConnectionStateReason.USER_DISCONNECTED,
                #     NM.VpnConnectionStateReason.DEVICE_DISCONNECTED,
                #     NM.VpnConnectionStateReason.SERVICE_STOPPED,
                #     NM.VpnConnectionStateReason.IP_CONFIG_INVALID,
                #     NM.VpnConnectionStateReason.SERVICE_START_FAILED,
                #     NM.VpnConnectionStateReason.CONNECTION_REMOVED,
                # ]
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        elif state == NM.VpnConnectionState.DISCONNECTED:
            if reason in [NM.VpnConnectionStateReason.USER_DISCONNECTED]:
                stop_local_agent()
                self._notify_subscribers_threadsafe(
                    events.Disconnected(EventContext(connection=self, reason=reason))
                )
            elif reason is NM.VpnConnectionStateReason.DEVICE_DISCONNECTED:
                stop_local_agent()
                self._notify_subscribers_threadsafe(
                    events.DeviceDisconnected(EventContext(connection=self, reason=reason))
                )
            else:
                # reason in [
                #     NM.VpnConnectionStateReason.UNKNOWN,
                #     NM.VpnConnectionStateReason.NONE,
                #     NM.VpnConnectionStateReason.SERVICE_STOPPED,
                #     NM.VpnConnectionStateReason.IP_CONFIG_INVALID,
                #     NM.VpnConnectionStateReason.CONNECT_TIMEOUT,
                #     NM.VpnConnectionStateReason.SERVICE_START_TIMEOUT,
                #     NM.VpnConnectionStateReason.SERVICE_START_FAILED,
                #     NM.VpnConnectionStateReason.NO_SECRETS,
                #     NM.VpnConnectionStateReason.LOGIN_FAILED,
                #     NM.VpnConnectionStateReason.CONNECTION_REMOVED,
                # ]
                self._notify_subscribers_threadsafe(
                    events.UnexpectedError(EventContext(connection=self, reason=reason))
                )
        else:
            logger.debug("Ignoring VPN state change: %s", state.value_name)

    async def update_settings(self, settings: Settings):
        """Update features on the active agent connection."""
        await super().update_settings(settings)
        if self._agent_listener.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            await self._request_connection_features(settings.features)


class OpenVPNTCP(OpenVPN):
    """Creates a OpenVPNTCP connection."""
    protocol = "openvpn-tcp"
    ui_protocol = "OpenVPN (TCP)"

    @classmethod
    def _get_priority(cls):
        return 1

    @classmethod
    def _validate(cls):
        # FIX ME: This should do a validation to ensure that NM can be used
        return True


class OpenVPNUDP(OpenVPN):
    """Creates a OpenVPNUDP connection."""
    protocol = "openvpn-udp"
    ui_protocol = "OpenVPN (UDP)"

    @classmethod
    def _get_priority(cls):
        return 1

    @classmethod
    def _validate(cls):
        # FIX ME: This should do a validation to ensure that NM can be used
        return True



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import re
from importlib import metadata
from typing import Optional

import time
import random
import os as sys_os
import json
from dataclasses import asdict
import distro
from packaging.version import Version
from proton.vpn import logging

logger = logging.getLogger(__name__)


class Serializable:  # pylint: disable=missing-class-docstring
    """Utility class for dataclasses."""
    def to_json(self) -> str:  # pylint: disable=missing-function-docstring
        return json.dumps(asdict(self))

    def to_dict(self) -> dict:  # pylint: disable=missing-function-docstring
        return asdict(self)

    @classmethod
    def from_dict(cls, dict_data: dict) -> 'Serializable':  # noqa: E501 pylint: disable=missing-function-docstring
        return cls._deserialize(dict_data)

    @classmethod
    def from_json(cls, data: str) -> 'Serializable':  # pylint: disable=missing-function-docstring
        dict_data = json.loads(data)
        return cls._deserialize(dict_data)

    @staticmethod
    def _deserialize(dict_data: dict) -> 'Serializable':
        raise NotImplementedError


class RefreshCalculator:
    """Calculates refresh times based on a set refresh randomness value."""
    def __init__(
        self,
        refresh_interval: int,
        refresh_randomness_in_percentage: float = None
    ):
        """
        The variable refresh_randomness_in_percentage will be used to create a
        deviation from original refresh value.
        Ie: 0.22 == 22% variation, so if we make request every 3h they will
        happen with random deviation between 0% and 22% from the base 3h value.
        """
        self._refresh_interval = refresh_interval
        self._refresh_randomness = refresh_randomness_in_percentage or 0.22

    @staticmethod
    def get_is_expired(expiration_time: float) -> bool:
        """Returns if data has expired"""
        current_time = time.time()
        return current_time > expiration_time

    @staticmethod
    def get_seconds_until_expiration(expiration_time: float) -> float:
        """
        Amount of seconds left until the client configuration is considered
        outdated and should be fetched again from the REST API.
        """
        seconds_left = expiration_time - time.time()
        return seconds_left if seconds_left > 0 else 0

    @staticmethod
    def get_expiration_time(
        refresh_interval: int,
        refresh_randomness: float = None,
        start_time: float = None
    ) -> float:  # noqa: E501 pylint: disable=missing-function-docstring
        """Returns the expiration time based on either a defined start time or current time."""
        start_time = start_time if start_time is not None else time.time()
        refresh_calculator = RefreshCalculator(refresh_interval, refresh_randomness)

        return start_time + refresh_calculator.get_refresh_interval_in_seconds()

    def get_refresh_interval_in_seconds(self) -> float:  # noqa pylint: disable=missing-function-docstring
        return self._refresh_interval * self._generate_random_component()

    def _generate_random_component(self):
        return 1 + self._refresh_randomness * (2 * random.random() - 1)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311


async def rest_api_request(session, route, **api_request_kwargs):  # noqa: E501 pylint: disable=missing-function-docstring
    logger.info(f"'{route}'", category="api", event="request")
    response = await session.async_api_request(
        route, **api_request_kwargs
    )
    logger.info(f"'{route}'", category="api", event="response")
    return response


def to_semver_build_metadata_format(value: Optional[str]) -> Optional[str]:
    """
    Formats the input value in a format that complies with
    semver's build metadata specs (https://semver.org/#spec-item-10).
    """
    if value is None:
        return None

    value = value.replace("_", "-")
    # Any character not allowed by semver's build metadata suffix
    # specs (https://semver.org/#spec-item-10) is removed.
    value = re.sub(r"[^a-zA-Z0-9\-]", "", value)
    return value


def get_desktop_environment() -> str:
    """Returns the current desktop environment"""
    return sys_os.environ.get('XDG_CURRENT_DESKTOP', "Unknown DE")


def get_distro_variant() -> str:
    """Returns the current distro environment"""
    distro_variant = distro.os_release_attr('variant')
    return f"; {distro_variant}" if distro_variant else ""


def get_distro_version() -> str:
    """Returns the string containing the distro version:
    ie:
     - Fedora: "39"/"40"
    """
    return distro.version()


def semver_from_pep440(pep440_version: str) -> str:
    """
    Converts a PEP440 version to a semver version.

    Disclaimers:
    - It assumes the PEP440 version contains the major, minor, micro triplet (e.g. 1.2.3).
    - Date-based releases are not supported (e.g. 2023.05).
    - Post release segments are not supported, since semver doesn't allow them.

    https://peps.python.org/pep-0440
    https://semver.org
    """
    ver = Version(pep440_version)

    # Even though PEP440 doesn't require it, our versions always contain
    # the major, minor, and micro triplet.
    result = f"{ver.major}.{ver.minor}.{ver.micro}"

    if ver.pre is not None:
        prerelease_mappings = {
            "a": "alpha",
            "b": "beta",
            "rc": "rc"
        }
        result += f"-{prerelease_mappings[ver.pre[0]]}.{ver.pre[1]}"

    if ver.dev is not None:
        result += f"-dev.{ver.dev}"

    if ver.local is not None:
        result += f"+{ver.local}"

    return result


def get_core_api_semver_version() -> str:
    """
    Converts the PEP440 version of this python package to the equivalent semver version.
    """
    return semver_from_pep440(metadata.version("proton-vpn-api-core"))


def generate_os_string() -> str:
    """Returns a string which contains information such as the distro, desktop environment
    and distro variant if it exists"""
    return f"{distro.id()} ({get_desktop_environment()}{get_distro_variant()})"



"""All the assets the app uses are available in this module.


Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List, Dict
from enum import Enum
from dataclasses import dataclass, field


class SplitTunnelingMode(Enum):
    """Enum for split tunneling mode.
    """
    EXCLUDE = "exclude"
    INCLUDE = "include"


@dataclass
class SplitTunnelingConfig:
    """Contains split tunneling data.
    """
    mode: SplitTunnelingMode = SplitTunnelingMode.EXCLUDE
    app_paths: List[str] = field(default_factory=list)
    ip_ranges: List[str] = field(default_factory=list)

    @staticmethod
    def from_dict(data: dict):
        """Generates `SplitTunnelingConfig` from regular python dict.

        Args:
            data (dict): the dict containing the necessary information

        Returns:
            SplitTunnelingConfig: new `SplitTunnelingConfig`
        """
        if not data:
            return SplitTunnelingConfig()

        return SplitTunnelingConfig(
            mode=SplitTunnelingMode(data.get("mode")),
            app_paths=data.get("app_paths", []),
            ip_ranges=data.get("ip_ranges", [])
        )

    def to_dict(self) -> dict:
        """Converts actual object to dict.

        Returns:
            dict: current object in dict
        """
        return {
            "mode": self.mode.value,
            "app_paths": self.app_paths,
            "ip_ranges": self.ip_ranges
        }


@dataclass
class SplitTunneling:
    """Config that is used for split tunneling
    """
    enabled: bool = False
    mode: SplitTunnelingMode = SplitTunnelingMode.EXCLUDE
    config_by_mode: dict[SplitTunnelingMode, SplitTunnelingConfig] = field(
        default_factory=lambda: {  # nosemgrep: python.lang.maintainability.return.return-not-in-function  # pylint: disable=line-too-long  # noqa: E501
            SplitTunnelingMode.EXCLUDE.value: SplitTunnelingConfig(mode=SplitTunnelingMode.EXCLUDE),
            SplitTunnelingMode.INCLUDE.value: SplitTunnelingConfig(mode=SplitTunnelingMode.INCLUDE)
        }
    )

    @property
    def exclude(self) -> SplitTunnelingConfig:
        """Returns the split tunneling config for the exclude mode."""
        return self.config_by_mode.get(
            SplitTunnelingMode.EXCLUDE.value,
            SplitTunnelingConfig(mode=SplitTunnelingMode.EXCLUDE)
        )

    @property
    def include(self) -> SplitTunnelingConfig:
        """Returns the split tunneling config for the include mode."""
        return self.config_by_mode.get(
            SplitTunnelingMode.INCLUDE.value,
            SplitTunnelingConfig(mode=SplitTunnelingMode.INCLUDE)
        )

    @staticmethod
    def from_dict(data: dict) -> SplitTunneling:
        """Generates `SplitTunneling` from regular python dict.

        Args:
            data (dict): the dict containing the necessary information

        Returns:
            SplitTunneling: new `SplitTunneling`
        """
        if not data:
            return SplitTunneling()

        raw_data = data.get("config_by_mode")

        if not raw_data:
            config_by_mode = {
                SplitTunnelingMode.EXCLUDE.value:
                    SplitTunnelingConfig(mode=SplitTunnelingMode.EXCLUDE),
                SplitTunnelingMode.INCLUDE.value:
                    SplitTunnelingConfig(mode=SplitTunnelingMode.INCLUDE)
            }
        else:
            config_by_mode = {
                SplitTunnelingMode(k).value: SplitTunnelingConfig.from_dict(v)
                for k, v in raw_data.items()
            }

        mode = SplitTunnelingMode(data.get("mode", SplitTunnelingMode.EXCLUDE.value))

        return SplitTunneling(
            enabled=data.get("enabled", False),
            mode=mode,
            config_by_mode=config_by_mode
        )

    def get_config(self) -> SplitTunnelingConfig:
        """Returns the split tunneling config for the currently selected mode.

        Returns:
            SplitTunnelingConfig: the split tunneling object
        """
        return self.config_by_mode[self.mode.value]

    def to_dict(self) -> Dict[str, object]:
        """Converts actual object to dict.

        Returns:
            dict: current object in dict
        """
        config_by_mode: dict[str, dict[str, str]] = \
            {k: v.to_dict() for k, v in self.config_by_mode.items()}

        return {
            "enabled": self.enabled,
            "mode": self.mode.value,
            "config_by_mode": config_by_mode
        }



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import functools
import itertools
import random
import time
from enum import Enum
from typing import Callable, Generator, Iterable, List, Optional

from proton.vpn import logging
from proton.vpn.session.dataclasses.servers import Country
from proton.vpn.session.exceptions import ServerNotFoundError, ServerListDecodeError
from proton.vpn.session.servers.types import LogicalServer, \
    TierEnum, ServerFeatureEnum, ServerLoad

logger = logging.getLogger(__name__)

UNIX_EPOCH = "Thu, 01 Jan 1970 00:00:00 GMT"


class PersistenceKeys(Enum):
    """JSON Keys used to persist the ServerList to disk."""
    LOGICALS = "LogicalServers"  # pylint: disable=R0902
    EXPIRATION_TIME = "ExpirationTime"
    LOADS_EXPIRATION_TIME = "LoadsExpirationTime"
    LAST_MODIFIED_TIME = "LastModifiedTime"
    USER_TIER = "MaxTier"
    STATUS_TOKEN = "StatusToken"  # nosec B105


class ServerList:  # pylint: disable=R0902, R0904
    """
    Server list model class.
    """

    LOGICALS_REFRESH_INTERVAL = 3 * 60 * 60  # 3 hours
    LOADS_REFRESH_INTERVAL = 15 * 60  # 15 minutes in seconds
    REFRESH_RANDOMNESS = 0.22  # +/- 22%

    """
    Wrapper around a list of logical servers.
    """
    def __init__(
            self,
            user_tier: TierEnum,
            logicals: Optional[List[LogicalServer]] = None,
            expiration_time: Optional[int] = None,
            loads_expiration_time: Optional[int] = None,
            index_servers: bool = True,
            last_modified_time: Optional[str] = None,
            status_token: Optional[str] = None
    ):  # pylint: disable=too-many-arguments
        self._user_tier = user_tier
        self._logicals = logicals or []
        self._expiration_time = expiration_time if expiration_time is not None\
            else ServerList.get_expiration_time()
        self._loads_expiration_time = loads_expiration_time if loads_expiration_time is not None\
            else ServerList.get_loads_expiration_time()
        self._last_modified_time = last_modified_time or ServerList.get_epoch_time()

        if index_servers:
            self._logicals_by_id, self._logicals_by_name = self._build_indexes(logicals)
        else:
            self._logicals_by_id = None
            self._logicals_by_name = None

        self._status_token = status_token

    @staticmethod
    def _build_indexes(logicals):
        logicals_by_id = {}
        logicals_by_name = {}

        for logical_server in logicals:
            logicals_by_id[logical_server.id] = logical_server
            logicals_by_name[logical_server.name.upper()] = logical_server

        return logicals_by_id, logicals_by_name

    @property
    def user_tier(self) -> TierEnum:
        """Tier of the user that requested the server list."""
        return self._user_tier

    @property
    def logicals(self) -> List[LogicalServer]:
        """The internal list of logical servers."""
        return self._logicals

    @property
    def expiration_time(self) -> float:
        """The expiration time of the server list as a unix timestamp."""
        return self._expiration_time

    @property
    def expired(self) -> bool:
        """
        Returns whether the server list expired, and therefore should be
        downloaded again, or not.
        """
        return time.time() > self._expiration_time

    @property
    def loads_expiration_time(self) -> float:
        """The expiration time of the server loads as a unix timestamp."""
        return self._loads_expiration_time

    @property
    def loads_expired(self) -> bool:
        """
        Returns whether the server list loads expired, and therefore should be
        updated, or not.
        """
        return time.time() > self._loads_expiration_time

    @property
    def last_modified_time(self) -> str:
        """The time at which the server list was fetched."""
        return self._last_modified_time

    @property
    def status_token(self) -> Optional[str]:
        """The token used to recover the status endpoint"""
        return self._status_token

    def update(self, server_loads: List[ServerLoad]):
        """Updates the server list with new server loads."""
        try:
            for server_load in server_loads:
                try:
                    logical_server = self.get_by_id(server_load.id)
                    logical_server.update(server_load)
                except ServerNotFoundError:
                    # Currently /vpn/loads returns some extra servers not returned by /vpn/logicals
                    logger.debug(f"Logical server was not found for update: {server_load}")
        finally:
            # If something unexpected happens when updating the server loads
            # it's safer to always update the loads expiration time to avoid
            # clients potentially retrying in a loop.
            self._loads_expiration_time = ServerList.get_loads_expiration_time()

    @property
    def seconds_until_expiration(self) -> float:
        """
        Amount of seconds left until the server list is considered outdated.

        The server list is considered outdated when
         - the full server list expires or
         - the server loads expire,
         whatever is the closest.
        """
        secs_until_full_expiration = max(self.expiration_time - time.time(), 0)
        secs_until_loads_expiration = max(self.loads_expiration_time - time.time(), 0)
        return min(secs_until_full_expiration, secs_until_loads_expiration)

    def get_by_id(self, server_id: str) -> LogicalServer:
        """
        :returns: the logical server with the given id.
        :raises ServerNotFoundError: if there is not a server with a matching id.
        """
        if self._logicals_by_id is None:
            raise RuntimeError("The server list was not indexed.")
        try:
            return self._logicals_by_id[server_id]
        except KeyError as error:
            raise ServerNotFoundError(
                f"The server with {server_id=} was not found"
            ) from error

    def get_by_name(self, name: str) -> LogicalServer:
        """
        :returns: the logical server with the given name.
        :raises ServerNotFoundError: if there is not a server with a matching name.
        """
        if self._logicals_by_name is None:
            raise RuntimeError("The server list was not indexed.")
        upper_case_name = name.upper()
        try:
            return self._logicals_by_name[upper_case_name]
        except KeyError as error:
            raise ServerNotFoundError(
                f"The server with {name=} was not found"
            ) from error

    @staticmethod
    def get_fastest_server(servers: Iterable[LogicalServer]) -> Optional[LogicalServer]:
        """
        :returns: Fastest server from the passed LogicalServer iterable
        """
        return min(servers, key=lambda s: s.score, default=None)

    @staticmethod
    def get_available_servers(
            servers: Iterable[LogicalServer],
            user_tier: TierEnum
    ) -> Generator[LogicalServer]:
        """
        :returns: Generator producing available servers
        from the passed LogicalServer iterable
        """
        return (
            server for server in servers
            if (
                server.enabled
                and server.tier <= user_tier
            )
        )

    @staticmethod
    def _compact_features(features: List[ServerFeatureEnum]) -> ServerFeatureEnum:
        return functools.reduce(lambda f1, f2: f1 | f2, features, 0)

    @staticmethod
    def get_servers_with_features(
            servers: Iterable[LogicalServer],
            request_features: ServerFeatureEnum = 0,
            exclude_features: ServerFeatureEnum = 0,
    ) -> Generator[LogicalServer]:
        """
        :returns: Generator producing servers matching/excluding specified features
        from the passed LogicalServer iterable
        """
        return (
            s for s in servers
            if (ServerList._compact_features(s.features) & exclude_features) == 0
            and (ServerList._compact_features(s.features) & request_features) == request_features
        )

    @staticmethod
    def get_servers_in_country_code(
            servers: Iterable[LogicalServer],
            country_code: str
    ) -> Generator[LogicalServer]:
        """
        :returns: Generator producing servers in the requested country
        from the passed LogicalServer iterable
        """
        return (
            server for server in servers
            if server.exit_country.lower() == country_code.lower()
        )

    @staticmethod
    def get_servers_in_city(
            servers: Iterable[LogicalServer],
            city_name: str
    ) -> Generator[LogicalServer]:
        """
        :returns: Generator producing servers in the requested city
        from the passed LogicalServer iterable
        """
        return (
            server for server in servers
            if server.city.lower() == city_name.lower()
        )

    def get_fastest_in_country(self, country_code: str) -> LogicalServer:
        """
        :returns: the fastest server for the specified country code and the tiers
        the user has access to.
        """
        country_servers = ServerList.get_servers_in_country_code(self.logicals, country_code)
        available_country_servers =\
            ServerList.get_available_servers(country_servers, self.user_tier)
        available_country_servers =\
            ServerList.get_servers_with_features(
                available_country_servers,
                exclude_features=ServerFeatureEnum.SECURE_CORE | ServerFeatureEnum.TOR
            )
        fastest_available_server = ServerList.get_fastest_server(available_country_servers)

        if not fastest_available_server:
            raise ServerNotFoundError("No server available in the current tier")

        return fastest_available_server

    def get_fastest_in_city(self, city_name: str) -> LogicalServer:
        """
        :returns: the fastest server in the specified city and the tiers
        the user has access to.
        """
        city_servers = ServerList.get_servers_in_city(self.logicals, city_name)
        available_city_servers = ServerList.get_available_servers(city_servers, self.user_tier)
        available_city_servers =\
            ServerList.get_servers_with_features(
                available_city_servers,
                exclude_features=ServerFeatureEnum.SECURE_CORE | ServerFeatureEnum.TOR
            )
        fastest_available_server = ServerList.get_fastest_server(available_city_servers)

        if not fastest_available_server:
            raise ServerNotFoundError("No server available in the current tier")

        return fastest_available_server

    def get_fastest(self) -> LogicalServer:
        """:returns: the fastest server in the tiers the user has access to."""
        available_servers = ServerList.get_available_servers(self.logicals, self.user_tier)
        available_servers =\
            ServerList.get_servers_with_features(
                available_servers,
                exclude_features=ServerFeatureEnum.SECURE_CORE | ServerFeatureEnum.TOR
            )
        fastest_available_server = ServerList.get_fastest_server(available_servers)

        if not fastest_available_server:
            raise ServerNotFoundError("No server available in the current tier")

        return fastest_available_server

    def group_by_country(self, cities: bool = False) -> List[Country]:
        """
        Returns the servers grouped by country.

        :param cities: whether to group the servers by cities as well.

        Before grouping the servers, they are sorted alphabetically by
        country name and server name.

        :return: The list of countries, each of them containing the servers
        in that country.
        """
        if cities:
            self.sort(sort_servers_by_country_and_city_and_enabled_and_load)
        else:
            self.sort()

        return [
            Country(country_code, list(country_servers))
            for country_code, country_servers in itertools.groupby(
                self.logicals, lambda server: server.exit_country.lower()
            )
        ]

    @classmethod
    def _generate_random_component(cls):
        # 1 +/- 0.22*random  # nosec B311
        return 1 + cls.REFRESH_RANDOMNESS * (2 * random.random() - 1)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311

    @classmethod
    def get_expiration_time(cls, start_time: int = None):
        """Returns the unix time at which the whole server list expires."""
        start_time = start_time if start_time is not None else time.time()
        return start_time + cls._get_refresh_interval_in_seconds()

    @classmethod
    def get_epoch_time(cls) -> str:
        """Returns the default fetch time in UTC which is the unix epoch.

        In the format of If-Modified-Since header which is
            <day-name>, <day> <month> <year> <hour>:<minute>:<second> GMT
        """
        return UNIX_EPOCH

    @classmethod
    def _get_refresh_interval_in_seconds(cls):
        return cls.LOGICALS_REFRESH_INTERVAL * cls._generate_random_component()

    @classmethod
    def get_loads_expiration_time(cls, start_time: int = None):
        """
        Generates the unix time at which the server loads will expire.
        """
        start_time = start_time if start_time is not None else time.time()
        return start_time + cls.get_loads_refresh_interval_in_seconds()

    @classmethod
    def get_loads_refresh_interval_in_seconds(cls) -> float:
        """
        Calculates the amount of seconds to wait before the server list should
        be fetched again from the REST API.
        """
        return cls.LOADS_REFRESH_INTERVAL * cls._generate_random_component()

    @classmethod
    def from_dict(
            cls, data: dict
    ):
        """
        :returns: the server list built from the given dictionary.
        """
        try:
            user_tier = data[PersistenceKeys.USER_TIER.value]
            logicals = [LogicalServer(logical_dict) for logical_dict in data["LogicalServers"]]
        except KeyError as error:
            raise ServerListDecodeError("Error building server list from dict") from error

        expiration_time = data.get(
            PersistenceKeys.EXPIRATION_TIME.value,
            cls.get_expiration_time()
        )
        loads_expiration_time = data.get(
            PersistenceKeys.LOADS_EXPIRATION_TIME.value,
            cls.get_loads_expiration_time()
        )

        last_modified_time = data.get(PersistenceKeys.LAST_MODIFIED_TIME.value,
                                      ServerList.get_epoch_time())

        status_token = data.get(PersistenceKeys.STATUS_TOKEN.value, None)

        return ServerList(
            user_tier=user_tier,
            logicals=logicals,
            expiration_time=expiration_time,
            loads_expiration_time=loads_expiration_time,
            last_modified_time=last_modified_time,
            status_token=status_token
        )

    def to_dict(self) -> dict:
        """:returns: the server list instance converted back to a dictionary."""
        return {
            PersistenceKeys.LOGICALS.value: [logical.to_dict() for logical in self.logicals],
            PersistenceKeys.EXPIRATION_TIME.value: self.expiration_time,
            PersistenceKeys.LOADS_EXPIRATION_TIME.value: self.loads_expiration_time,
            PersistenceKeys.LAST_MODIFIED_TIME.value: self.last_modified_time,
            PersistenceKeys.USER_TIER.value: self._user_tier,
            PersistenceKeys.STATUS_TOKEN.value: self._status_token
        }

    def __len__(self):
        return len(self.logicals)

    def __iter__(self):
        yield from self.logicals

    def __getitem__(self, item):
        return self.logicals[item]

    def sort(self, key: Callable = None):
        """See List.sort()."""
        key = key or sort_servers_alphabetically_by_country_and_server_name
        self.logicals.sort(key=key)


def sort_servers_alphabetically_by_country_and_server_name(server: LogicalServer) -> str:
    """
    Returns the comparison key used to sort servers alphabetically,
    first by exit country name and then by server name.

    If the server name is in the form of COUNTRY-CODE#NUMBER, then NUMBER
    is padded with zeros to be able to sort the server name in natural sort
    order.
    """
    country_name = server.exit_country_name
    server_name = server.name or ""
    server_name = server_name.lower()
    if "#" in server_name:
        # Pad server number with zeros to achieve natural sorting
        server_name = f"{server_name.split('#')[0]}#" \
                      f"{server_name.split('#')[1].zfill(10)}"

    return f"{country_name}__{server_name}"


def sort_servers_by_country_and_city_and_enabled_and_load(server: LogicalServer) -> str:
    """
    Returns the comparison key used to sort servers by country name, city name,
    whether they are enabled or not, and load.
    """
    return (server.exit_country_name, server.city, 0 if server.enabled else 1, server.load)



"""
Module for Kill Switch based on Network Manager.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code

from typing import Optional, TYPE_CHECKING

import subprocess  # nosec B404:blacklist # nosemgrep: gitlab.bandit.B404

from proton.vpn.killswitch.interface import KillSwitch
from proton.vpn.backend.networkmanager.killswitch.default.killswitch_connection_handler\
    import KillSwitchConnectionHandler
from proton.vpn import logging

if TYPE_CHECKING:
    from proton.vpn.connection import VPNServer


logger = logging.getLogger(__name__)


class NMKillSwitch(KillSwitch):
    """
    Kill Switch implementation using NetworkManager.

    A dummy Network Manager connection is created to block all non-VPN traffic.

    The way it works is that the dummy connection blocking non-VPN traffic is
    added with a lower priority than the VPN connection but with a higher
    priority than the other network manager connections. This way, the routing
    table uses the dummy connection for any traffic that does not go to the
    primary VPN connection.
    """

    def __init__(self, ks_handler: KillSwitchConnectionHandler = None):
        self._ks_handler = ks_handler or KillSwitchConnectionHandler()
        super().__init__()

    async def enable(
            self, vpn_server: Optional["VPNServer"] = None, permanent: bool = False
    ):  # noqa
        """Enables general kill switch."""
        # The full KS blocks all traffic except the one going to an already
        # existing VPN interface.
        await self._ks_handler.add_full_killswitch_connection(permanent)

        # If the routed KS is already enabled then it needs to be removed.
        # There is no way to just update it with the new VPN server IP.
        await self._ks_handler.remove_routed_killswitch_connection()

        if not vpn_server:
            return

        # The routed KS blocks all traffic except the one going to the specified VPN server IP.
        await self._ks_handler.add_routed_killswitch_connection(vpn_server.server_ip, permanent)

        # At this point the full KS is removed to allow establishing the new VPN connection
        # to the specified server IP.
        await self._ks_handler.remove_full_killswitch_connection()

    async def disable(self):
        """Disables general kill switch."""
        await self._ks_handler.remove_full_killswitch_connection()
        await self._ks_handler.remove_routed_killswitch_connection()

    async def enable_ipv6_leak_protection(self, permanent: bool = False):
        """Enables IPv6 kill switch."""
        await self._ks_handler.add_ipv6_leak_protection()

    async def disable_ipv6_leak_protection(self):
        """Disables IPv6 kill switch."""
        await self._ks_handler.remove_ipv6_leak_protection()

    @staticmethod
    def _get_priority() -> int:
        return 100

    @staticmethod
    def _validate():
        try:
            KillSwitchConnectionHandler().is_network_manager_running  # noqa pylint: disable=expression-not-assigned disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
        except (ModuleNotFoundError, ImportError):
            logger.error("NetworkManager is not running.")
            return False

        # libnetplan0 is the first version that is present in Ubuntu 22.04. In Ubuntu 24.04
        # the package name changes to libnetplan1, and it's not compatible with this kill
        # switch implementation when IPv6 is disabled via the ipv6.disabled kernel option.
        try:
            result = subprocess.run(
                ["/usr/bin/apt", "show", "libnetplan1"],
                capture_output=True,
                check=True, shell=False
            )  # nosec B603:subprocess_without_shell_equals_true
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass
        else:
            stdout_decoded = result.stdout.decode("utf8").split("\n")
            for package_info_line in stdout_decoded:
                if package_info_line.startswith("Version: 1.0.0"):
                    logger.warning(
                        "Kill switch is not compatible with libnetplan1 v1.0.0. "
                        "Please upgrade libnetplan1 package to v1.1.1"
                    )
                    break

        return True



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=C0413
import asyncio
from getpass import getpass
from threading import Event
from typing import Iterator, Optional

from proton.vpn.session import fido2_handler
from proton.session import Session
from proton.session.api import Fido2AssertionParameters, Fido2Assertion
from proton.vpn.session.exceptions import (
    SecurityKeyError, Fido2NotSupportedError,
    SecurityKeyNotFoundError, InvalidSecurityKeyError,
    SecurityKeyPINInvalidError, SecurityKeyPINNotSetError,
    SecurityKeyTimeoutError
)
from proton.vpn.session.u2f_interaction import UserInteraction


from fido2.hid import CtapHidDevice    # pylint: disable=wrong-import-order
# pylint: disable=no-name-in-module
from fido2.client import (             # pylint: disable=wrong-import-order
    Fido2Client, ClientError, UserInteraction as Fido2UserInteraction
)
from fido2.ctap import CtapError       # pylint: disable=wrong-import-order
from fido2.ctap2.pin import ClientPin  # pylint: disable=wrong-import-order


class Fido2UserInteractionAdaptor(Fido2UserInteraction):
    """
    Wraps a UserInteraction object to provide the Fido2UserInteraction
    interface.

    See UserInteraction in fido2.client for details.
    """
    def __init__(self, user_interaction: UserInteraction):
        """Initialize the adaptor with the given UserInteraction object."""
        self._user_interaction = user_interaction

    def prompt_up(self) -> None:
        """Called when the authenticator is awaiting a user presence check."""
        self._user_interaction.prompt_up()

    def request_pin(
        self, permissions: ClientPin.PERMISSION, rp_id: Optional[str]
    ) -> Optional[str]:
        """Called when the client requires a PIN from the user"""
        return self._user_interaction.request_pin(permissions, rp_id)

    def request_uv(
        self, permissions: ClientPin.PERMISSION, rp_id: Optional[str]
    ) -> bool:
        """Called when the client is about to request UV from the user."""
        return self._user_interaction.request_uv(permissions, rp_id)


class U2FKeys:
    """Manage U2F keys."""

    def list_devices(self) -> Iterator[CtapHidDevice]:
        """List all connected FIDO2 devices."""
        return CtapHidDevice.list_devices()

    async def scan_keys_and_get_assertion(
            self,
            session: Session,
            user_interaction: Optional[UserInteraction] = None,
            cancel_assertion: Optional[Event] = None
    ) -> Fido2Assertion:
        """
        Select a FIDO2 client and get an assertion from it.
        :param session: user session.
        :param user_interaction: optional object handling any required user interaction.
        :param cancel_assertion: optional event to be able to cancel the ongoing assertion.
        :returns: the generated FIDO2 assertion.
        """
        if not session.supports_fido2:
            raise Fido2NotSupportedError("Session does not support FIDO2 authentication")

        origin = "https://" + session.supports_fido2.rp_id
        fido2_clients = [
            # pylint: disable=unexpected-keyword-arg
            fido2_handler.create_client(
                device,
                origin,
                Fido2UserInteractionAdaptor(user_interaction)
                if user_interaction else Fido2UserInteraction()
            )
            for device in self.list_devices()
        ]

        if not fido2_clients:
            raise SecurityKeyNotFoundError("No security key found")

        if len(fido2_clients) == 1:
            selected_client = fido2_clients[0]
        else:
            user_interaction.request_key_selection()
            selected_client = await self._touch_key_to_use(fido2_clients)

        assertion_parameters = session.supports_fido2
        cancel_assertion = cancel_assertion or Event()
        return await self.get_assertion_from_client(
            selected_client, assertion_parameters, cancel_assertion
        )

    async def get_assertion_from_client(
            self,
            client: Fido2Client,
            assertion_parameters: Fido2AssertionParameters,
            cancel_assertion: Event
    ) -> Fido2Assertion:
        """Get an assertion from the given FIDO2 client."""
        options = fido2_handler.create_options(assertion_parameters)
        try:
            assertion_selection = await asyncio.to_thread(
                client.get_assertion, options, cancel_assertion
            )
        except ClientError as error:
            if error.code == ClientError.ERR.DEVICE_INELIGIBLE:
                raise InvalidSecurityKeyError("The security key is not eligible") from error

            if error.code == ClientError.ERR.TIMEOUT:
                raise SecurityKeyTimeoutError("The security key operation timed out") from error

            if error.code == ClientError.ERR.CONFIGURATION_UNSUPPORTED:
                raise SecurityKeyPINNotSetError(
                    "The security key doesn't have a PIN set but the server requires it"
                ) from error

            if (
                error.code == ClientError.ERR.BAD_REQUEST
                and isinstance(error.cause, CtapError)
                and error.cause.ERR.PIN_INVALID
            ):
                raise SecurityKeyPINInvalidError(
                    "The security key PIN provided is not valid"
                ) from error

            raise SecurityKeyError("An error occurred with the security key") from error

        except OSError as error:
            # if the key is removed while the client is waitint for it to be touched we get:
            # OSError: [Errno 19] No such device
            raise SecurityKeyNotFoundError("The security key could not be accessed") from error

        except Exception as error:
            raise SecurityKeyError("An error occcurred with the security key") from error

        return fido2_handler.create_from_client_assertion(assertion_selection)

    async def _touch_key_to_use(self, fido2_clients: list[Fido2Client]) -> Fido2Client:
        cancel_key_selection = Event()

        tasks = [
            asyncio.create_task(asyncio.to_thread(
                self._client_selection, client, cancel_key_selection
            ))
            for client in fido2_clients
        ]

        done_tasks, _ = await asyncio.wait(tasks)
        results = [task.result() for task in done_tasks]
        selected_client = [client for client in results if client is not None].pop()

        return selected_client

    def _client_selection(
            self, client: Fido2Client, cancel_client_selection: Event
    ) -> Optional[Fido2Client]:
        try:
            # Block until user touches the key or event is set
            client.selection(cancel_client_selection)
        except ClientError as error:
            if error.code != ClientError.ERR.TIMEOUT:
                raise
            return None

        # Cancel other client selections
        cancel_client_selection.set()

        # Return the selected client
        return client


class CLIUserInteraction:
    """
    Provides user interaction via CLI to the Fido2 client.

    The interface currently follows the fido2.client.UserInteraction interface,
    adding some methods to it.
    """

    def prompt_up(self) -> None:
        """Called when the authenticator is awaiting a user presence check."""
        print("If your security key has a button or a gold disc, tap it now to authenticate.")

    def request_pin(
        self, *_args, **_kwargs
    ) -> Optional[str]:
        """Called when the client requires a PIN from the user.

        Should return a PIN, or None/Empty to cancel."""
        return getpass("Introduce your PIN and press enter: ")

    def request_uv(self, *_args, **_kwargs) -> bool:
        """Called when the client is about to request UV from the user.

        Should return True if allowed, or False to cancel."""
        return True

    def request_key_selection(self):
        """Called when multiple keys are found and the user needs to select one
        by touching it."""
        print(
            "Multiple security keys were detected. "
            "If your security key has a button or a gold disc, tap it now to select it."
        )


async def main():
    """Example usage of the U2FKeys class."""

    session = Session()
    username = input("Enter your username: ")
    password = getpass("Enter your password: ")
    await session.async_authenticate(username=username, password=password)

    if not session.authenticated:
        raise RuntimeError("Authentication failed")

    if not session.needs_twofa:
        raise RuntimeError("Session does not need 2FA")

    print("Scanning for keys...")
    manager = U2FKeys()
    assertion = await manager.scan_keys_and_get_assertion(session, CLIUserInteraction())

    print("FIDO2 assertion:", assertion)
    result = await session.async_validate_2fa_fido2(assertion)
    if result:
        print("2FA successful, session is now fully authenticated.")
    else:
        print("2FA failed.")
    await session.async_logout()


if __name__ == "__main__":
    asyncio.run(main())



"""
This modules contains the classes that communicate with NetworkManager.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=duplicate-code
# pylint: disable=duplicate-code
import re
import subprocess  # nosec blacklist # nosemgrep: gitlab.bandit.B404
import asyncio
import concurrent.futures

from proton.vpn import logging
from proton.vpn.backend.networkmanager.killswitch.wireguard.nmclient import (
    NMClient, GatewayNotFoundError
)
from proton.vpn.backend.networkmanager.killswitch.wireguard.killswitch_connection import (
    KillSwitchConnection, KillSwitchGeneralConfig, KillSwitchIPConfig
)

logger = logging.getLogger(__name__)


def _get_connection_id(permanent: bool, ipv6: bool = False):
    if ipv6:
        return f"pvpn-killswitch-ipv6{'-perm' if permanent else ''}"

    return f"pvpn-killswitch{'-perm' if permanent else ''}"


def _get_interface_name(permanent: bool, ipv6: bool = False):
    if ipv6:
        return f"ipv6leakintrf{'1' if permanent else '0'}"

    return f"pvpnksintrf{'1' if permanent else '0'}"


async def _wrap_future(future: concurrent.futures.Future, timeout=10):
    """Wraps a concurrent.future.Future object in an asyncio.Future object."""
    return await asyncio.wait_for(
        asyncio.wrap_future(future, loop=asyncio.get_running_loop()),
        timeout=timeout
    )


class KillSwitchConnectionHandler:
    """Kill switch connection management."""

    def __init__(
            self, nm_client: NMClient = None,
            config: KillSwitchGeneralConfig = None,
            server_ip: str = None
    ):
        self._nm_client = nm_client
        self._config = config
        self._ipv4_ks_settings = KillSwitchIPConfig(
            addresses=["100.85.0.1/24"],
            dns=["0.0.0.0"],  # nosec hardcoded_bind_all_interfaces
            dns_priority=-1400,
            gateway="100.85.0.1",
            ignore_auto_dns=True,
            route_metric=98,
            routes=[]  # accept/block all routes.
        )
        self._ipv6_ks_settings = KillSwitchIPConfig(
            addresses=["fdeb:446c:912d:08da::/64"],
            dns=["::1"],
            dns_priority=-1400,
            gateway="fdeb:446c:912d:08da::1",
            ignore_auto_dns=True,
            route_metric=95
        )
        self._server_ip = server_ip

    @property
    def nm_client(self):
        """Returns the NetworkManager client."""
        if self._nm_client is None:
            self._nm_client = NMClient()

        return self._nm_client

    @property
    def is_network_manager_running(self) -> bool:
        """Returns if the Network Manager daemon is running or not."""
        return self.nm_client.get_nm_running()

    @property
    def is_connectivity_check_enabled(self) -> bool:
        """Returns if connectivity_check property is enabled or not."""
        return self.nm_client.connectivity_check_get_enabled()

    async def add_kill_switch_connection(self, permanent: bool):
        """Adds a dummy connection that swallows all traffic it receives.

        This dummy connection has more priority than an ethernet/wifi
        interface but with less priority than the VPN connection."""
        await self._ensure_connectivity_check_is_disabled()

        general_config = self._config or KillSwitchGeneralConfig(
            human_readable_id=_get_connection_id(permanent),
            interface_name=_get_interface_name(permanent)
        )

        connection = self.nm_client.get_active_connection(
            conn_id=general_config.human_readable_id
        )

        if connection:
            logger.debug("Kill switch was already present.")
            return

        kill_switch = KillSwitchConnection(
            general_config,
            ipv4_settings=self._ipv4_ks_settings,
            ipv6_settings=self._ipv6_ks_settings,
        )
        await _wrap_future(
            self.nm_client.add_connection_async(kill_switch.connection, save_to_disk=permanent)
        )
        logger.debug(f"{'Permanent' if permanent else 'Non-permanent'} kill switch added.")
        await self._remove_connection(
            connection_id=_get_connection_id(permanent=not permanent)
        )
        logger.debug(f"{'Non-permanent' if permanent else 'Permanent'} kill switch removed.")

    async def add_vpn_server_route(self, server_ip: str):
        """Add route to allow outgoing traffic to the specified IP."""
        await self._ensure_connectivity_check_is_disabled()

        if not self.nm_client.is_monitoring_network_config_changes():
            self._start_monitoring_network_config_changes()

        devices = self.nm_client.get_physical_devices()
        for device in devices:
            try:
                await _wrap_future(
                    self.nm_client.add_route_to_device(
                        device, new_server_ip=server_ip,
                        old_server_ip=self._server_ip
                    )
                )
            except GatewayNotFoundError as error:
                logger.warning(
                    "Wireguard connection cannot use interface "
                    f"'{device.get_iface()}: {error}'")
                continue

            # The new route doesn't seem to be available straight away.
            # For this reason, the routing table is polled until the route has been added.
            await self._wait_for_vpn_server_route(
                server_ip, device.get_iface(), found=True
            )

        self._server_ip = server_ip

    async def remove_vpn_server_route(self):
        """
        Remove a previously added VPN server route.
        If the route is not found then nothing happens.
        """
        if self.nm_client.is_monitoring_network_config_changes():
            self.nm_client.stop_monitoring_network_config_changes()

        if not self._server_ip:
            return

        devices = self.nm_client.get_physical_devices()
        for device in devices:
            await _wrap_future(
                self.nm_client.remove_route_from_device(device, self._server_ip)
            )
            # The route doesn't seem to be removed straight away.
            # For this reason, the routing table is polled until the route has been removed.
            await self._wait_for_vpn_server_route(self._server_ip, device.get_iface(), found=False)

        self._server_ip = None

    @staticmethod
    async def _run_ip_route_command():
        def run():
            return subprocess.run(  # nosec subprocess_without_shell_equals_true
                ["/usr/sbin/ip", "route"], capture_output=True, encoding="utf-8", check=True
            )

        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, run)

    @classmethod
    async def _wait_for_vpn_server_route(
            cls, server_ip: str, interface_name: str, found: bool = True
    ):
        server_route = f"{server_ip} via .* dev {interface_name} .*"
        for delay in [0.5, 0.5, 1, 1, 2]:
            result = await cls._run_ip_route_command()

            if bool(re.search(server_route, result.stdout)) is found:
                return

            await asyncio.sleep(delay)

        raise TimeoutError(
            f"Error waiting for server route to be {'added' if found else 'removed'}"
        )

    def _start_monitoring_network_config_changes(self):
        loop = asyncio.get_running_loop()

        def on_active_connection_changed(active_connection):
            device = active_connection.get_devices()[0]
            gateway = active_connection.get_ip4_config().get_gateway()
            logger.info(f"Interface {device.get_iface()} switched to gateway {gateway}")
            future = self.nm_client.add_route_to_device(device, self._server_ip, gateway)
            future.add_done_callback(lambda f: loop.call_soon_threadsafe(f.result))

        self.nm_client.start_monitoring_network_config_changes(on_active_connection_changed)

    async def add_ipv6_leak_protection(self):
        """Adds IPv6 kill switch to prevent IPv6 leaks while using IPv4."""
        await self._ensure_connectivity_check_is_disabled()

        connection_id = _get_connection_id(permanent=False, ipv6=True)
        connection = self.nm_client.get_active_connection(
            conn_id=connection_id)

        if connection:
            logger.debug("IPv6 leak protection already present.")
            return

        interface_name = _get_interface_name(permanent=False, ipv6=True)
        general_config = KillSwitchGeneralConfig(
            human_readable_id=connection_id,
            interface_name=interface_name
        )

        kill_switch = KillSwitchConnection(
            general_config,
            ipv4_settings=None,
            ipv6_settings=self._ipv6_ks_settings,
        )

        await _wrap_future(
            self.nm_client.add_connection_async(kill_switch.connection, save_to_disk=False)
        )
        logger.debug("IPv6 leak protection added.")

    async def remove_killswitch_connection(self):
        """Removes full kill switch connection."""
        logger.debug("Removing full kill switch...")
        if self._config:
            await self._remove_connection(self._config.human_readable_id)
        else:
            await self._remove_connection(_get_connection_id(permanent=True))
            await self._remove_connection(_get_connection_id(permanent=False))
        logger.debug("Full kill switch removed.")

    async def remove_ipv6_leak_protection(self):
        """Removes IPv6 kill switch connection."""
        logger.debug("Removing IPv6 leak protection...")
        await self._remove_connection(_get_connection_id(permanent=False, ipv6=True))
        logger.debug("IP6 leak protection removed.")

    async def _remove_connection(self, connection_id: str):
        connection = self.nm_client.get_connection(
            conn_id=connection_id)

        logger.debug(f"Attempting to remove {connection_id}: {connection}")

        if not connection:
            logger.debug(f"There was no {connection_id} to remove")
            return

        await _wrap_future(self.nm_client.remove_connection_async(connection))

    async def _ensure_connectivity_check_is_disabled(self):
        if self.is_connectivity_check_enabled:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            await _wrap_future(self.nm_client.disable_connectivity_check())
            logger.info("Network connectivity check was disabled.")



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


class VPNSessionNotLoadedError(Exception):
    """
    Data from the current VPN session was accessed before it was loaded.
    """


class VPNAccountDecodeError(ValueError):
    """The VPN account could not be deserialized."""


class VPNCertificateError(Exception):
    """
    Base class for certificate errors.
    """


class VPNCertificateExpiredError(VPNCertificateError):
    """
    VPN Certificate is available but is expired.
    """


class VPNCertificateNeedRefreshError(VPNCertificateError):
    """
    VPN Certificate is available but needs to be refreshed because
    is close to expiration.
    """


class VPNCertificateFingerprintError(VPNCertificateError):
    """
    VPN Certificate and private key fingerprint are not matching.
    A new keypair should be generated and the corresponding certificate
    should be fetched from our REST API.
    """


class ServerListDecodeError(ValueError):
    """The server list could not be parsed."""


class ServerNotFoundError(Exception):
    """
    The specified server could not be found in the server list.
    """


class ClientConfigDecodeError(ValueError):
    """The client configuration could not be parsed."""


class SecurityKeyError(Exception):
    """Base exception for security key errors."""


class Fido2NotSupportedError(SecurityKeyError):
    """Raised when FIDO2 authentication is not available on the current session."""


class SecurityKeyNotFoundError(SecurityKeyError):
    """Raised when no security key is found."""


class InvalidSecurityKeyError(SecurityKeyError):
    """Raised when the security key is invalid or cannot be used."""


class SecurityKeyTimeoutError(SecurityKeyError):
    """Raised when the security key operation times out."""


class SecurityKeyPINNotSetError(SecurityKeyError):
    """
    Raised when the FIDO 2 server requires a PIN to be set
    on the security key, but the user didn't set it.
    """


class SecurityKeyPINInvalidError(SecurityKeyError):
    """
    Raised when the security key has a PIN set but the one the
    user provided is not correct.
    """



"""
This module contains then logic related to storing forwarded port to file.

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import pathlib
from proton.vpn.connection import states
from proton.utils.environment import VPNExecutionEnvironment

DEFAULT_FORWARDED_PORT_FILEPATH = os.path.join(
    VPNExecutionEnvironment().path_runtime, "forwarded_port"
)


class PortForwardFileHandler:  # pylint: disable=too-few-public-methods
    """Takes care of file creating/deleting and writing port to file."""

    def __init__(self, filepath: str = DEFAULT_FORWARDED_PORT_FILEPATH):
        self._filepath = pathlib.Path(filepath)

    def on_state_change_update_port(self, state: states.State):
        """Handles the new state.
        If the state is `Connected`, it writes the forwarded port to the file.
        If the state is not `Connected`, it removes the port from the file.

        Args:
            state (State): newly received state object.
        """
        if isinstance(state, states.Connected) and state.forwarded_port:
            self._write_port_to_file(state.forwarded_port)
        elif isinstance(state, (states.Connected, states.Disconnected)):
            self._remove_port_from_file()

    def _write_port_to_file(self, port: int):
        """Writes port to file."""
        with open(self._filepath, mode="w", encoding="utf8") as file:
            file.write(str(port))

    def _remove_port_from_file(self):
        """Removes the port from the file."""
        with open(self._filepath, mode="w", encoding="utf8") as file:
            file.write("")



"""
Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import inspect
import time
from asyncio import CancelledError

from dataclasses import dataclass
from typing import Optional, Coroutine, List, Callable


@dataclass
class RunAgain:
    """Object to be returned by a task to be run again after a certain amount of time."""
    delay_in_ms: int

    @staticmethod
    def after_seconds(seconds: float):
        """Returns a RunAgain object to be run after a certain amount of seconds."""
        return RunAgain(delay_in_ms=int(seconds * 1000))


@dataclass
class TaskRecord:
    """Record with details of the task to be executed and when."""
    id: int  # pylint: disable=invalid-name
    timestamp: float
    async_function: Callable[[], Coroutine]
    background_task: Optional[asyncio.Task] = None


class Scheduler:
    """
    Task scheduler.

    The goal of this implementation is to improve the accuracy of the built-in scheduler
    when the system is suspended/resumed. The built-in scheduler does not take into account
    the time the system has been suspended after a task has been scheduled to run after a
    certain amount of time. In this case, the clock is paused and then resumed.

    The way this implementation workarounds this issue is by keeping a record of tasks to
    be executed and the timestamp at which they should be executed. Then it periodically
    checks the lists for any tasks that should be executed and runs them.
    """

    def __init__(self, check_interval_in_ms: int = 10_000):
        self._check_interval_in_ms = check_interval_in_ms
        self._error_callback = None
        self._last_task_id: int = 0
        self._task_list: List[TaskRecord] = []
        self._scheduler_task: Optional[asyncio.Task] = None

    def set_error_callback(self, error_callback: Callable[[Exception], None] = None):
        """Sets the error callback to be called when an error occurs while executing a task."""
        self._error_callback = error_callback

    def unset_error_callback(self):
        """Unsets the error callback."""
        self._error_callback = None

    @property
    def task_list(self):
        """Returns the list of tasks currently scheduled."""
        return self._task_list

    @property
    def is_started(self):
        """Returns whether the scheduler has been started or not."""
        return self._scheduler_task is not None

    @property
    def number_of_remaining_tasks(self):
        """Returns the number of remaining tasks to be executed."""
        return len([record for record in self._task_list if not record.background_task])

    def get_tasks_ready_to_fire(self) -> List[TaskRecord]:
        """
        Returns the tasks that are ready to fire, that is the tasks with a timestamp lower or
        equal than the current unix time."""
        now = time.time()
        return list(filter(
            lambda record: record.timestamp <= now and not record.background_task,
            self._task_list
        ))

    def start(self):
        """Starts the scheduler."""
        if self.is_started:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            raise RuntimeError("Scheduler was already started.")

        self._scheduler_task = asyncio.create_task(self._run_periodic_task_list_check())

    async def stop(self):
        """Stops the scheduler and discards all remaining tasks."""
        if self.is_started:    # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            self._scheduler_task.cancel()

            for record in self._task_list:
                if record.background_task:
                    record.background_task.cancel()
            self._task_list = []

            await self.wait_for_shutdown()
            self._scheduler_task = None

    async def wait_for_shutdown(self, timeout=1):
        """Waits for the scheduler to be stopped."""
        if self.is_started:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            try:
                await asyncio.wait_for(self._scheduler_task, timeout)
            except CancelledError:
                pass

    def run_soon(self, async_function: Callable[[], Coroutine]) -> int:
        """
        Runs the coroutine as soon as possible.
        :returns: the scheduled task id.
        """
        return self.run_after(0, async_function)

    def run_after(
            self, delay_in_seconds: float, async_function: Callable[[], Coroutine]
    ) -> int:
        """
        Runs the coroutine after a delay specified in seconds.
        :returns: the scheduled task id.
        """
        return self.run_at(time.time() + delay_in_seconds, async_function)

    def run_at(
            self, timestamp: float, async_function: Callable[[], Coroutine]
    ) -> int:
        """
        Runs the task at the specified timestamp.
        :returns: the scheduled task id.
        """
        if not inspect.iscoroutinefunction(async_function):
            raise ValueError("A coroutine function was expected.")

        self._last_task_id += 1

        record = TaskRecord(
            id=self._last_task_id,
            timestamp=timestamp,
            async_function=async_function
        )
        self._task_list.append(record)

        return record.id

    def cancel_task(self, task_id):
        """Cancels a task to be executed given its task id."""
        for task in self._task_list:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.correctness.list-modify-iterating.list-modify-while-iterate
            if task.id == task_id:
                if task.background_task:
                    task.background_task.cancel()
                else:
                    self._task_list.remove(task)
                break  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.correctness.list-modify-iterating.list-modify-while-iterate

    async def _run_periodic_task_list_check(self):
        while True:
            self.run_tasks_ready_to_fire()
            await asyncio.sleep(self._check_interval_in_ms / 1000)

    def run_tasks_ready_to_fire(self):
        """
        Runs the tasks ready to be executed, that is the tasks with a timestamp lower or equal
        than the current unix time, and removes them from the list.
        """
        tasks_ready_to_fire = self.get_tasks_ready_to_fire()

        # Run the tasks that are ready to be run.
        for task_record in tasks_ready_to_fire:
            task = asyncio.create_task(task_record.async_function())
            task_record.background_task = task
            task.add_done_callback(self._on_task_done)

    def _on_task_done(self, task: asyncio.Task):
        # Get the task record associated with the task.
        task_record = next(filter(lambda record: record.background_task == task, self._task_list))

        result = None
        try:
            # Bubble up exceptions, if any.
            result = task.result()
        except CancelledError:
            # CancelledError is raised when the task is cancelled.
            pass
        except Exception as exc:  # pylint: disable=broad-except
            self._task_list.remove(task_record)
            if not self._error_callback:
                raise exc
            self._error_callback(exc)
            return

        if isinstance(result, RunAgain):
            # if the task record is to be run again then it's rescheduled.
            task_record.timestamp = time.time() + result.delay_in_ms / 1000
            task_record.background_task = None
        else:
            self._task_list.remove(task_record)



"""
This module defines the classes holding the necessary configuration to establish
a VPN connection.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import ipaddress
import tempfile
import os

from proton.utils.environment import ExecutionEnvironment


class VPNConfiguration:
    """Base VPN configuration."""
    PROTOCOL = None
    EXTENSION = None

    def __init__(self, vpnserver, vpncredentials, settings):
        self._configfile = None
        self._configfile_enter_level = None
        self._vpnserver = vpnserver
        self._vpncredentials = vpncredentials
        self._settings = settings

    def __enter__(self):
        # We create the configuration file when we enter,
        # and delete it when we exit.
        # This is a race free way of having temporary files.

        if self._configfile is None:
            self._delete_existing_configuration()
            # NOTE: we should try to keep filename length
            # below 15 characters, including the prefix.
            self._configfile = tempfile.NamedTemporaryFile(
                dir=self.__base_path, delete=False,
                prefix='pvpn', suffix=self.EXTENSION, mode='w'
            )
            self._configfile.write(self.generate())
            self._configfile.close()
            self._configfile_enter_level = 0

        self._configfile_enter_level += 1

        return self._configfile.name

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._configfile is None:
            return

        self._configfile_enter_level -= 1
        if self._configfile_enter_level == 0:
            os.unlink(self._configfile.name)
            self._configfile = None

    def _delete_existing_configuration(self):
        for file in self.__base_path:
            if file.endswith(f".{self.EXTENSION}"):
                os.remove(os.path.join(self.__base_path, file))

    def generate(self) -> str:
        """
        Generates the configuration file content.
        """
        raise NotImplementedError

    @property
    def __base_path(self):
        return ExecutionEnvironment().path_runtime

    @staticmethod
    def cidr_to_netmask(cidr) -> str:
        """Returns the subnet netmask from the CIDR."""
        subnet = ipaddress.IPv4Network(f"0.0.0.0/{cidr}")
        return str(subnet.netmask)

    @staticmethod
    def is_valid_ipv4(ip_address) -> bool:
        """Returns True if the specified ip address is a valid IPv4 address,
        and False otherwise."""
        try:
            ipaddress.ip_address(ip_address)
        except ValueError:
            return False

        return True



"""
Local Agent module.


Copyright (c) 2024 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import proton.vpn.local_agent
import proton.vpn.logging

from proton.vpn.local_agent import (  # pylint: disable=no-name-in-module, import-error
    AgentConnector, AgentConnection, Status,
    State, Reason, ReasonCode, AgentFeatures,
    LocalAgentError, ExpiredCertificateError,
    PolicyAPIError, SyntaxAPIError, APIError,
    ConnectionDetails
)
from .listener import AgentListener

# Initialize logging for the local agent module, this forwards the rust
# logging to the python logging module.
proton.vpn.local_agent.init_logger(proton.vpn.logging.getLogger)

__all__ = [
    "AgentConnector", "AgentConnection", "Status",
    "State", "Reason", "ReasonCode", "AgentFeatures",
    "LocalAgentError", "ExpiredCertificateError",
    "PolicyAPIError", "SyntaxAPIError", "APIError",
    "ConnectionDetails", "AgentListener"
]
